"""
     esbmtk: A general purpose Earth Science box model toolkit
     Copyright (C), 2020 Ulrich G. Wortmann

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
import numpy as np
import numpy.typing as npt
import matplotlib.pyplot as plt
import logging
import typing as tp
from esbmtk import Q_
import functools
from numba import njit

if tp.TYPE_CHECKING:
    from esbmtk import Flux, Model, Connection, Connect, ExternalFunction

np.set_printoptions(precision=4)
# declare numpy types
NDArrayFloat = npt.NDArray[np.float64]


class ScaleError(Exception):
    def __init__(self, message):
        message = f"\n\n{message}\n"
        super().__init__(message)


def rmtree(f) -> None:
    """Delete file, of file is directorym delete all files in

    :param f: pathlib path object

    """
    if f.is_file():
        f.unlink()
    else:
        for child in f.iterdir():
            rmtree(child)
        f.rmdir()


def phc(c: float) -> float:
    """Calculate concentration as pH. c can be a number or numpy array

    Parameters
    ----------
    c : float
        H+ concentration

    Returns
    -------
    float
        pH value

    """
    import numpy as np

    pH: float = -np.log10(c)
    return pH


def debug(func):
    """Print the function signature and return value"""

    @functools.wraps(func)
    def wrapper_debug(*args, **kwargs):
        args_repr = [f"{float(repr(a)):.2e}\n" for a in args]  # 1
        kwargs_repr = [f"{k}={v!r}" for k, v in kwargs.items()]  # 2
        signature = ", ".join(args_repr + kwargs_repr)  # 3
        print(f"Calling {func.__name__}\n({signature})")
        value = func(*args, **kwargs)
        value = [f"{v:.2e}" for v in value]
        print(f"\n{func.__name__!r} returned\n {value!r}\n")  # 4
        return value

    return wrapper_debug


def get_reservoir_reference(k: str, M: Model) -> tuple:
    """Get Species and Reservoir handles

    Parameters
    ----------
    k : str
        with the initial flux name, e.g., M_F_A_db_DIC
    M : Model
        Model handle
    Returns
    -------
    tuple
        Reservoir/Connection, Species
    Raises
    ------
    ValueError
        If reservoir_name is not of type Reservoir/Ggroup or Connection

    """
    from esbmtk import Reservoir, ReservoirGroup, Connection, Connect
    from esbmtk import GasReservoir, Species

    key_list = k[2:].split(".")  # get model, reservoir & species name

    if len(key_list) == 3:  # ReservoirGroup
        model_name, reservoir_name, species_name = key_list

        if hasattr(M, reservoir_name):
            obj = getattr(M, reservoir_name)
        else:
            raise ValueError(f"{reservoir_name} is not in the model")

        if k[0:2] == "R_":
            reservoir = obj
        elif k[0:2] == "F_":
            if isinstance(obj, Reservoir | ReservoirGroup):
                reservoir = obj
            elif isinstance(obj, Connection | Connect):
                reservoir = obj
                if isinstance(reservoir.source, GasReservoir):
                    species_name = obj.sink.name
                else:
                    species_name = obj.source.name
            else:
                raise ValueError(
                    f"{obj.name} must be Reservoir or Connection", f"not {type(obj)}"
                )

    elif len(key_list) == 2:  # (Gas)Reservoir
        model_name, reservoir_name = key_list
        reservoir = getattr(M, reservoir_name)
    else:
        raise ValueError("kl should look like this F_M.CO2_At")

    species = getattr(M, species_name)
    if not isinstance(species, Species):
        breakpoint()
    return reservoir, species


def register_new_flux(rg, dict_key, dict_value) -> list:
    """Register a new flux object with a Reservoir or Connection instance

    Parameters
    ----------
    rg : Reservoir | ReservoirGroup
        instance to register with
    dict_key : str
        E.g., "M.A_db.DIC"
    dict_value : str
        id value, e.g., "db_cs2"

    Returns
    -------
    list
        list of Flux instances

    """
    from .esbmtk import Reservoir, Flux, Source
    from .extended_classes import ReservoirGroup

    if isinstance(rg, ReservoirGroup):
        spn = dict_key.split(".")[-1]
        sp = getattr(rg.mo, spn)
        reg = getattr(rg, sp.name)
    elif isinstance(rg, Reservoir):
        spn = dict_key.split(".")[-1]
        sp = getattr(rg.mo, spn)
        reg = rg
    else:
        print(f" type(rg) = {type(rg)}")
        raise NotImplementedError
   
    if not hasattr(reg, "source"):
        setattr(reg, "source", sp.name)

    ro = list()
    f = Flux(
        name=dict_value,
        species=sp,
        rate=0,
        register=reg,
        ftype="computed",
        isotopes=reg.isotopes,
    )

    ro.append(f)
    reg.lof.append(f)  # register flux
    reg.ctype = "ignore"  # FIXME why is this here?

    return ro


def register_new_reservoir(r, sp, v):
    """Register a new reservoir"""
    from .esbmtk import Reservoir

    rt = Reservoir(  # create new reservoir
        name=sp.name,
        species=sp,
        concentration=f"{v} mol/kg",
        register=r,
        volume=r.volume,
        rtype="computed",
    )
    r.lor.append(rt)
    return [rt]


def register_return_values(ef: ExternalFunction, rg) -> None:
    """Register the return values of an external function instance

    Parameters
    ----------
    ec : ExternalFunction
        ExternalFunction Instance
    rg : ReservoirGroup | Reservoir
        The Resevoir or Reservoirgroup the external function is
        associated with

    Raises
    ------
    ValueError
        If the return value type is undefined

    Check the return values of external function instances,
    and create the necessary reservoirs, fluxes, or connections
    if they are missing.

    These fluxes are not associated with a connection Object
    so we register the source/sink relationship with the
    reservoir they belong to.

    This fails for GasReservoirs since they can have a 1:many
    relatioship. The below is a terrible hack, it would be
    better to express this with several connection
    objects, rather than overloading the source attribute of the
    GasReservoir class.
    """
    from esbmtk import Reservoir, ReservoirGroup, Flux, Connection, Connect
    from esbmtk import Sink, SinkGroup

    M = rg.mo
    # go through each entry in ec.return_values
    for line in ef.return_values:
        if isinstance(line, dict):
            dict_key = next(iter(line))  # get first key
            dict_value = line[dict_key]
            if dict_key[:2] == "F_":  # is flux
                key_str = dict_key[2:].split(".")[1]
                if hasattr(M, key_str):
                    o = getattr(M, key_str)
                    if isinstance(o, Flux):
                        o: list = [o]
                    elif isinstance(o, Connection | Connect):
                        o: list = [getattr(o, "_F")]  # get flux handle
                    elif isinstance(o, Reservoir | ReservoirGroup):
                        o: list = register_new_flux(rg, dict_key[2:], dict_value)
                    elif isinstance(o, Sink | SinkGroup):
                        o: list = register_new_flux(rg, dict_key[2:], dict_value)
                    else:
                        raise ValueError(f"No recipie for {type(o)}")
                else:
                    raise ValueError(f"{key_str} is not part of the Model definition")

            elif dict_key[:2] == "R_":  # is reservoir
                if dict_key[2:] in M.lor:
                    o: list = [getattr(M, dict_key)]
                else:
                    r, sp = get_reservoir_reference(dict_key, M)
                    o: list = register_new_reservoir(r, sp, dict_value)

            elif dict_key[:2] == "C_":  # is connection
                raise NotImplementedError
            else:
                raise ValueError(f"{dict_key[0:2]} is not defined")

        elif isinstance(line, Reservoir):
            dict_value.ef_results = True
            o = [dict_value]
        else:
            raise ValueError("This should not happen")

        ef.lro.extend(o)  # add to list of returned Objects


def summarize_results(M: Model) -> dict():
    """Summarize all model results at t_max into a hirarchical
    dictionary, where values are accessed in the following way:

    results[basin_name][level_name][species_name]

    e.g., result["A"]["sb"]["O2"]

    """
    results = dict()

    for r in M.lor:
        species_name = r.name
        basin_name = r.register.name[:1]
        level_name = r.register.name[-2:]
        species_value = r.c[-1]

        if basin_name not in results:
            results[basin_name] = {level_name: {species_name: species_value}}
        elif level_name not in results[basin_name]:
            results[basin_name][level_name] = {species_name: species_value}
        elif species_name not in results[basin_name][level_name]:
            results[basin_name][level_name][species_name] = species_value

    return results


def find_matching_strings(s: str, fl: list[str]) -> bool:
    """test if all elements of fl occur in s. Return True if yes,
    otherwise False

    """
    return all(f in s for f in fl)


def insert_into_namespace(name, value, name_space=globals()):
    name_space[name] = value


def add_to(l, e):
    """
    add element e to list l, but check if the entry already exist. If so, throw
    exception. Otherwise add
    """

    if e not in l:  # if not present, append element
        l.append(e)


def get_plot_layout(obj):
    """Simple function which selects a row, column layout based on the number of
    objects to display.  The expected argument is a reservoir object which
    contains the list of fluxes in the reservoir

    """

    noo = 1 + sum(f.plot == "yes" for f in obj.lof)
    for _ in obj.ldf:
        noo += 1

    # noo = len(obj.lof) + 1  # number of objects in this reservoir
    logging.debug(f"{noo} subplots for {obj.n} required")

    size, geo = plot_geometry(noo)

    return size, geo


def plot_geometry(noo: int) -> tuple():
    """Define plot geometry based on number of objects to plot"""

    if noo < 2:
        geo = [1, 1]  # one row, one column
        size = [5, 3]  # size in inches
    elif 1 < noo < 3:
        geo = [2, 1]  # two rows, one column
        size = [5, 6]  # size in inches
    elif 2 < noo < 5:
        geo = [2, 2]  # two rows, two columns
        size = [10, 6]  # size in inches
    elif 4 < noo < 7:
        geo = [2, 3]  # two rows, three columns
        size = [15, 6]  # size in inches
    elif 6 < noo < 10:
        geo = [3, 3]  # two rows, three columns
        size = [15, 9]  # size in inches
    elif 9 < noo < 13:
        geo = [4, 3]  # two rows, three columns
        size = [15, 12]  # size in inches
    elif 12 < noo < 16:
        geo = [5, 3]  # two rows, three columns
        size = [15, 15]  # size in inches
    else:
        m = (
            "plot geometry for more than 15 fluxes is not yet defined"
            "Consider calling flux.plot individually on each flux in the reservoir"
        )
        raise ValueError(m)

    return size, geo


def list_fluxes(self, name, i) -> None:
    """
    Echo all fluxes in the reservoir to the screen
    """
    for f in self.lof:  # show the processes
        direction = self.lio[f.n]
        if direction == -1:
            t1 = "From:"
            t2 = "Outflux from"
        else:
            t1 = "To  :"
            t2 = "Influx to"

        for p in f.lop:
            p.describe()

    for f in self.lof:
        f.describe(i)  # print out the flux data


def show_data(self, **kwargs) -> None:
    """Print the 3 lines of the data starting with index

    Optional arguments:

    index :int = 0 starting index
    indent :int = 0 indentation
    """

    off: str = "  "

    index = 0 if "index" not in kwargs else kwargs["index"]
    ind: str = kwargs["indent"] * " " if "indent" in kwargs else ""


def set_y_limits(ax: plt.Axes, obj: any) -> None:
    """Prevent the display or arbitrarily small differences"""

    bottom, top = ax.get_ylim()
    if (top - bottom) < obj.display_precision:
        top = bottom + obj.display_precision
        bottom -= obj.display_precision
        ax.set_ylim(bottom, top)


def is_name_in_list(n: str, l: list) -> bool:
    """Test if an object name is part of the object list"""

    return any(e.full_name == n for e in l)


def get_object_from_list(name: str, l: list) -> any:
    """Match a name to a list of objects. Return the object"""

    match: bool = False
    for o in l:
        if o.full_name == name:
            r = o
            match = True

    if match:
        return r
    else:
        raise ValueError(f"Object = {o.full_name} has no matching flux {name}")


def sort_by_type(l: list, t: list, m: str) -> list:
    """divide a list by type into new lists. This function will return a
    list and it is up to the calling code to unpack the list

    l is list with various object types
    t is a list which contains the object types used for sorting
    m is a string for the error function
    """

    # from numbers import Number

    lc = l.copy()
    rl = []

    for ot in t:  # loop over object types
        a = []
        for e in l:  # loop over list elements
            if isinstance(e, ot):
                a.append(e)  # add to temporary list
                lc.remove(e)  # remove this element

        rl.append(a)  # save the temporary list to rl

    # at this point, all elements of lc should have been processed
    # if not, lc contains element which are of a different type
    if lc:
        raise TypeError(m)

    return rl


def get_object_handle(res: list, M: Model):
    """Test if the key is a global reservoir handle
    or exists in the model namespace

    :param res: list of strings, or reservoir handles
    :param M: Model handle
    """

    r_list: list = []

    if not isinstance(res, list):
        res = [res]
    for o in res:
        if o in M.dmo:  # is object known in global namespace
            r_list.append(M.dmo[o])
        elif o in M.__dict__:  # or does it exist in Model namespace
            r_list.append(getattr(M, o))
        else:
            print(f"{o} is not known for model {M.name}")
            raise ValueError(f"{o} is not known for model {M.name}")

    if len(r_list) == 1:
        r_list = r_list[0]

    return r_list


def split_key(k: str, M: any) -> tp.Union[any, any, str]:
    """split the string k with letters _to_, and test if optional
    id string is present

    """

    if "_to_" not in k:
        raise ValueError("Name must follow 'Source_to_Sink' format")

    source = k.split("_to_")[0]
    sinkandid = k.split("_to_")[1]
    if "@" in sinkandid:
        sink = sinkandid.split("@")[0]
        cid = sinkandid.split("@")[1]
    else:
        sink = sinkandid
        cid = ""

    sink = get_object_handle(sink, M)
    source = get_object_handle(source, M)

    return (source, sink, cid)


def make_dict(keys: list, values: list) -> dict:
    """Create a dictionary from a list and value, or from
    two lists

    """
    d = {}

    if isinstance(values, list):
        if len(values) == len(keys):
            d: dict = dict(zip(keys, values))
        else:
            raise ValueError("key and value list must be of equal length")
    else:
        values: list = [values] * len(keys)
        d: dict = dict(zip(keys, values))

    return d


def get_typed_list(data: list) -> list:
    tl = list()
    for x in data:
        tl.append(x)
    return tl


def create_reservoirs(box_dict: dict, ic_dict: dict, M: any) -> dict:
    """boxes are defined by area and depth interval here we use an ordered
    dictionary to define the box geometries. The next column is temperature
    in deg C, followed by pressure in bar
    the geometry is [upper depth datum, lower depth datum, area percentage]

    :param bn: dictionary with box parameters,

    e.g.::

     box_dict: dict = {  # name: [[geometry], T, P]
                 "sb": {"g": [0, 200, 0.9], "T": 20, "P": 5},
                 "ib": {"g": [200, 1200, 1], "T": 10, "P": 100},
                }

    :param ic: dictionary with species default values.

    ic is used to set up
    initial conditions. Here we use shortcut and use the same conditions
    in each box. If you need box specific initial conditions
    use the output of build_concentration_dicts as starting point, e.g.,::

      ic_dict: dict = { # species: concentration, Isotopes, delta, f_only
                   PO4: [Q_("2.1 * umol/liter"), False, 0, False],
                   DIC: [Q_("2.1 mmol/liter"), False, 0, False],
                   ALK: [Q_("2.43 mmol/liter"), False, 0, False],
               }

    :param M: Model object handle

    """

    from esbmtk import ReservoirGroup, build_concentration_dicts
    from esbmtk import SourceGroup, SinkGroup

    # loop over reservoir names
    for box_name, value in box_dict.items():
        # test key format
        if M.name in box_name:
            box_name = box_name.split(".")[1]

        keyword_dict: dict = build_concentration_dicts(ic_dict, box_name)

        if "ty" in value:  # type is given
            if value["ty"] == "Source":
                if "delta" in value:
                    SourceGroup(
                        name=box_name,
                        species=value["sp"],
                        delta=value["delta"],
                        isotopes=keyword_dict[box_name][1],
                        register=M,
                    )
                else:
                    SourceGroup(
                        name=box_name,
                        species=value["sp"],
                        register=M,
                        isotopes=keyword_dict[box_name][1],
                    )
            elif value["ty"] == "Sink":
                SinkGroup(
                    name=box_name,
                    species=value["sp"],
                    register=M,
                    isotopes=keyword_dict[box_name][1],
                )
            else:
                raise ValueError("'ty' must be either Source or Sink")

        else:  # create reservoirs
            rg = ReservoirGroup(
                name=box_name,
                geometry=value["g"],
                concentration=keyword_dict[box_name][0],
                isotopes=keyword_dict[box_name][1],
                delta=keyword_dict[box_name][2],
                seawater_parameters={
                    "temperature": value["T"],
                    "pressure": value["P"],
                    "salinity": value["S"],
                },
                register=M,
            )

    return keyword_dict


def build_concentration_dicts(cd: dict, bg: dict) -> dict:
    """Build a dict which can be used by create_reservoirs

    :param bg: dict where the box_names are dict keys.
    :param cd: dictionary

    with the following format::

        cd = {
             # species: [concentration, isotopes]
             PO4: [Q_("2.1 * umol/liter"), False],
             DIC: [Q_("2.1 mmol/liter"), False],
            }

    This function returns a new dict in the following format

    #  box_names: [concentrations, isotopes]
    d= {"bn": [{PO4: .., DIC: ..},{PO4:False, DIC:False}]}

    """

    if isinstance(bg, dict):
        box_names: list = bg.keys()
    elif isinstance(bg, list):
        box_names: list = bg
    elif isinstance(bg, str):
        box_names: list = [bg]
    else:
        raise ValueError("This should never happen")

    icd: dict = {}
    td1: dict = {}  # temp dictionary
    td2: dict = {}  # temp dictionary
    td3: dict = {}  # temp dictionary

    # create the dicts for concentration and isotopes
    for k, v in cd.items():
        td1[k] = v[0]
        td2[k] = v[1]
        td3[k] = v[2]

    # box_names: list = bg.keys()
    for bn in box_names:  # loop over box names
        icd[bn] = [td1, td2, td3]

    return icd


def calc_volumes(bg: dict, M: any, h: any) -> list:
    """Calculate volume contained in a given depth interval
    bg is a dictionary in the following format::

      bg={
          "hb": (0.1, 0, 200),
          "sb": (0.9, 0, 200),
         }

    where the key must be a valid box name, the first entry of the list denoted
    the areal extent in percent, the second number is upper depth limit, and last
    number is the lower depth limit.

    M must be a model handle
    h is the hypsometry handle

    The function returns a list with the corresponding volumes

    """

    # from esbmtk import hypsometry

    v: list = []  # list of volumes

    for v in bg.values():
        a = v[0]
        u = v[1]
        l = v[2]

        v.append(h.volume(u, l) * a)

    return v


def get_longest_dict_entry(d: dict) -> int:
    """Get length of each item in the connection dict"""
    l_length = 0  # length of  longest list
    p_length = 0  # length of single parameter
    nl = 0  # number of lists
    ll = []

    # we need to cover the case where we have two lists of different length
    # this happens if we have a long list of tuples with matched references,
    # as well as a list of species
    for v in d.values():
        if isinstance(v, list):
            nl = nl + 1
            if len(v) > l_length:
                l_length = len(v)
                ll.append(l_length)

        else:
            p_length = 1

    if nl > 1 and ll.count(l_length) != len(ll):
        raise ValueError("Mapping for multiple lists is not supported")

    if l_length > 0 and p_length == 0:
        case = 0  # Only lists present
    if l_length == 0 and p_length == 1:
        case = 1  # Only parameters present
    if l_length > 0 and p_length == 1:
        case = 2  # Lists and parameters present

    return case, l_length


def convert_to_lists(d: dict, l: int) -> dict:
    """expand mixed dict entries (i.e. list and single value) such
    that they are all lists of equal length

    """
    cd = d.copy()

    for k, v in cd.items():
        if not isinstance(v, list):
            p = [v for _ in range(l)]
            d[k] = p

    return d


def get_sub_key(d: dict, i: int) -> dict:
    """take a dict which has where the value is a list, and return the
    key with the n-th value of that list

    """

    return {k: v[i] for k, v in d.items()}


def expand_dict(d: dict, mt: str = "1:1") -> int:
    """Determine dict structure

    in case we have mutiple connections with mutiple species, the
    default action is to map connections to species (t = '1:1'). If
    you rather want to create mutiple connections (one for each
    species) in each connection set t = '1:N'

    """
    # loop over dict entries
    # ck = connection key
    # cd = connection dict

    r: dict = {}  # the dict we will return

    for ck, cd in d.items():  # loop over connections
        rd: dict = {}  # temp dict
        nd: dict = {}  # temp dict
        case, length = get_longest_dict_entry(cd)

        if isinstance(ck, tuple):
            # assume 1:1 mapping between tuple and connection parameters
            if mt == "1:1":
                # prep dictionaries
                if case == 0:
                    nd = cd
                elif case == 1:  # only parameters present. Expand for each tuple entry
                    length = len(ck)
                    nd = convert_to_lists(cd, length)
                elif case == 2:  # mixed list present, Expand list
                    nd = convert_to_lists(cd, length)

                # for each connection group in the tuple
                if length != len(ck):
                    message = (
                        f"The number of connection properties ({length})\n"
                        f"does not match the number of connection groups ({len(ck)})\n"
                        f"did you intend to do a 1:N mapping?"
                    )
                    raise ValueError(message)

                # map property dicts to connection group names
                i = 0
                for t in ck:
                    rd[t] = get_sub_key(nd, i)
                    i = i + 1

            elif mt == "1:N":  # apply each species to each connection
                if case == 0:
                    nd = cd
                elif case == 1:  # only parameters present. Expand for each tuple entry
                    length = len(ck)
                    nd = convert_to_lists(cd, length)
                elif case == 2:  # mixed list present, Expand list
                    nd = convert_to_lists(cd, length)

                for t in ck:  # apply the entire nd dict to all connections
                    rd[t] = nd
            else:
                raise ValueError(f"{mt} is not defined. must be '1:1' or '1:N'")

        else:
            if case in [0, 1]:  # only lists present, case 3
                nd = cd
            elif case == 2:  # list and parameters present case 4
                nd = convert_to_lists(cd, length)
            rd[ck] = nd

        # update the overall dict and move to the next entry
        # r |= rd
        r.update(rd)

    return r


def create_bulk_connections(ct: dict, M: Model, mt: int = "1:1") -> dict:
    """Create connections from a dictionary. The dict can have the following keys
    following format:

    mt = mapping type. See below for explanation

    # na: names, tuple or str. If lists, all list elements share the same properties
    # sp: species list or species
    # ty: type, str
    # ra: rate, Quantity
    # sc: scale, Number
    # re: reference, optional
    # al: alpha, optional
    # de: delta, optional
    # bp: bypass, see scale_with_flux
    # si: signal
    # mx: True, optional defaults to False. If set, it will create forward and backward fluxes (i.e. mixing)

    There are 6 different cases how to specify connections

    Case 1 One connection, one set of parameters
           ct1 = {"sb2hb": {"ty": "scale", 'ra'....}}

    Case 2 One connection, one set of instructions, one subset with multiple parameters
           This will be expanded to create connections for each species
           ct2 = {"sb2hb": {"ty": "scale", "sp": ["a", "b"]}}

    Case 3 One connection complete set of multiple characters. Similar to case 2,
           but now all parameters are given explicitly
           ct3 = {"sb2hb": {"ty": ["scale", "scale"], "sp": ["a", "b"]}}

    Case 4 Multiple connections, one set of parameters. This will create
           identical connection for "sb2hb" and  "ib2db"
           ct4 = {("sb2hb", "ib2db"): {"ty": "scale", 'ra': ...}}

    Case 5 Multiple connections, one subset of multiple set of parameters. This wil
          create a connection for species 'a' in sb2hb and with species 'b' in ib2db
           ct5 = {("sb2hb", "ib2db"): {"ty": "scale", "sp": ["a", "b"]}}

    Case 6 Multiple connections, complete set of parameters of multiple parameters
           Same as case 5, but now all parameters are specified explicitly
           ct6 = {("sb2hb", "ib2db"): {"ty": ["scale", "scale"], "sp": ["a", "b"]}}


    The default interpretation for cases 5 and 6 is that each list
    entry corresponds to connection. However, sometimes we want to
    create multiple connections for multiple entries. In this case
    provide the mt='1:N' parameter which will create a connection for
    each species in each connection group. See the below example.

    It is easy to shoot yourself in the foot. It is best to try the above first with
    some simple examples, e.g.,

    from esbmtk import expand_dict
    ct2 = {"sb2hb": {"ty": "scale", "sp": ["a", "b"]}}

    It is best to use the show_dict function to verify that your input
    dictionary produces the correct results!

    """

    from esbmtk import create_connection, expand_dict

    # expand dictionary into a well formed dict where each connection
    # has a fully formed entry
    c_ct = expand_dict(ct, mt=mt)

    # loop over dict entries and create the respective connections
    for k, v in c_ct.items():
        if isinstance(k, tuple):  # loop over names in tuple
            for c in k:
                create_connection(c, v, M)
        elif isinstance(k, str):
            create_connection(k, v, M)
        else:
            raise ValueError(f"{c} must be string or tuple")

    return c_ct


def create_connection(n: str, p: dict, M: Model) -> None:
    """called by create_bulk_connections in order to create a connection group
    It is assumed that all rates are in liter/year or mol per year.  This may
    not be what you want or need.

    :param n: a connection key.  if the mix flag is given interpreted as mixing
              a connection between sb and db and thus create connections in both
              directions
    :param p: a dictionary holding the connection properties
    :param M: the model handle
    """

    from esbmtk import ConnectionGroup, Q_

    # get the reservoir handles by splitting the key
    source, sink, cid = split_key(n, M)
    # create default connections parameters and replace with values in
    # the parameter dict if present.
    los = list(p["sp"]) if isinstance(p["sp"], list) else [p["sp"]]
    ctype = "None" if "ty" not in p else p["ty"]
    scale = 1 if "sc" not in p else p["sc"]
    rate = Q_("0 mol/a") if "ra" not in p else p["ra"]
    ref_flux = "None" if "re" not in p else p["re"]
    alpha = "None" if "al" not in p else p["al"]
    delta = "None" if "de" not in p else p["de"]
    cid = f"{cid}"
    bypass = "None" if "bp" not in p else p["bp"]
    species = "None" if "sp" not in p else p["sp"]
    signal = "None" if "si" not in p else p["si"]

    if isinstance(scale, Q_):
        if scale.check(["dimensionless"]):
            scale = scale.magnitude
        else:
            scale = scale.to(M.f_unit).magnitude

    # expand arguments
    ctype = make_dict(los, ctype)
    scale = make_dict(los, scale)  # get rate from dictionary
    rate = make_dict(los, rate)
    ref_flux = make_dict(los, ref_flux)
    alpha = make_dict(los, alpha)
    delta = make_dict(los, delta)
    bypass = make_dict(los, bypass)
    signal = make_dict(los, signal)

    name = f"{M.name}.CG_{source.name}_to_{sink.name}"
    if f"{name}" in M.lmo:  # Test if CG exists
        cg = getattr(M, name.split(".")[1])  # retriece CG object
        cg.add_connections(
            source=source,
            sink=sink,
            ctype=ctype,
            scale=scale,  # get rate from dictionary
            rate=rate,
            ref_flux=ref_flux,
            alpha=alpha,
            delta=delta,
            bypass=bypass,
            register=M,
            signal=signal,
            id=cid,  # get id from dictionary
        )
    else:  # Create New ConnectionGroup
        cg = ConnectionGroup(
            source=source,
            sink=sink,
            ctype=ctype,
            scale=scale,  # get rate from dictionary
            rate=rate,
            ref_flux=ref_flux,
            alpha=alpha,
            delta=delta,
            bypass=bypass,
            register=M,
            signal=signal,
            id=cid,  # get id from dictionary
        )


def get_name_only(o: any) -> any:
    """Test if item is an esbmtk type. If yes, extract the name"""

    from esbmtk import Flux, Reservoir, ReservoirGroup, Species

    return (
        o.full_name if isinstance(o, (Flux, Reservoir, ReservoirGroup, Species)) else o
    )


def get_simple_list(l: list) -> list:
    """return a list which only has the full name
    rather than all the object properties

    """

    return [get_name_only(e) for e in l]


def show_dict(d: dict, mt: str = "1:1") -> None:
    """show dict entries in an organized manner"""

    from esbmtk import expand_dict, get_simple_list, get_name_only

    ct = expand_dict(d, mt)
    for ck, cv in ct.items():
        for pk, pv in cv.items():
            x = get_simple_list(pv) if isinstance(pv, list) else get_name_only(pv)


def find_matching_fluxes(l: list, filter_by: str, exclude: str) -> list:
    """Loop over all reservoir in l, and extract the names of all fluxes
    which match the filter string. Return the list of names (not objects!)

    """

    lof: set = set()

    for r in l:
        for f in r.lof:
            if filter_by in f.full_name and exclude not in f.full_name:
                lof.add(f)

    return list(lof)


def reverse_key(key: str) -> str:
    """reverse a connection key e.g., sb2db@POM becomes db2sb@POM"""

    left = key.split("@")
    left = left[0]
    rs = left.split("_to_")
    r1 = rs[0]
    r2 = rs[1]

    return f"{r2}_to_{r1}"


def get_connection_keys(
    f_list: set,
    ref_id: str,
    target_id: str,
    inverse: bool,
    exclude: str,
) -> list[str]:
    """
    extract connection keys from set of flux names, replace ref_id with
    target_id so that the key can be used in create_bulk_connnections()

    :param f_list: a set with flux objects
    :param ref_id: string with the reference id
    :param target_id: string with the target_id
    :param inverse: Bool, optional, defaults to false

    :return cnc_l: a list of connection keys (str)

    The optional inverse parameter, can be used where in cases where the
    flux direction needs to be reversed, i.e., the returned key will not read
    sb2db@POM, but db2s@POM
    """

    cnc_l: list = []  # list of connection keys

    for f in f_list:
        # get connection and flux name
        fns = f.full_name.split(".")
        cnc = fns[1][3:]  # get key without leadinf C_
        if inverse:
            cnc = reverse_key(cnc)

        cnc.replace(ref_id, target_id)
        # create new cnc string
        cnc = f"{cnc}_to_{target_id}"
        cnc_l.append(cnc)

    return cnc_l


def gen_dict_entries(M: Model, **kwargs) -> tuple(tuple, list):
    """Find all fluxes that contain the reference string, and create a new
    Connection instance that connects the flux matching ref_id, with a flux
    matching target_id.  The function will a tuple containig the new connection
    keys that can be used by the create bulk_connection() function.  The second
    return value is a list containing the reference fluxes.

    The optional inverse parameter, can be used where in cases where the flux
    direction needs to be reversed, i.e., the returned key will not read
    sb_to_dbPOM, but db_to_sb@POM

    :param M: Model or list

    :param kwargs: keyword dictionary, known keys are ref_id, and raget_id, inverse

    :return f_list: List of fluxes that match ref_id

    :return k_tuples: tuple of connection keys

    """

    from esbmtk import Model

    ref_id = kwargs["ref_id"]
    target_id = kwargs["target_id"]
    inverse = kwargs.get("inverse", False)
    exclude_str = kwargs.get("exclude", "None")

    # find matching fluxes
    if isinstance(M, Model):
        f_list: list = find_matching_fluxes(
            M.loc,
            filter_by=ref_id,
            exclude=exclude_str,
        )
    elif isinstance(M, list):
        f_list: list = find_matching_fluxes(
            M,
            filter_by=ref_id,
            exclude=exclude_str,
        )
    else:
        raise ValueError(f"gen_dict_entries: M must be list or Model, not {type(M)}")

    k_tuples: list = get_connection_keys(
        f_list,
        ref_id,
        target_id,
        inverse,
        exclude_str,
    )

    return tuple(k_tuples), f_list


def build_ct_dict(d: dict, p: dict) -> dict:
    """build a connection dictionary from a dict containing connection
    keys, and a dict containing connection properties. This is most
    useful for connections which a characterized by a fixed rate but
    apply to many species. E.g., mixing fluxes in a complex model etc.

    """

    # a = {k: {"sc": v} | p for k, v in d.items()}
    a = {}
    for k, v in d.items():
        a[k] = p.copy()
        a[k]["sc"] = v

    return a


def get_string_between_brackets(s: str) -> str:
    """Parse string and extract substring between square brackets"""

    s = s.split("[")
    if len(s) < 2:
        raise ValueError(f"Column header {s} must include units in square brackets")

    s = s[1]

    s = s.split("]")

    if len(s) < 2:
        raise ValueError(f"Column header {s} must include units in square brackets")

    return s[0]


def check_for_quantity(quantity, unit):
    """check if keyword is quantity or string an convert as necessary

    Parameters
    ----------
    quantity : str | quantity | float | int
        e.g., "12 m/s", or 12,
    unit : str
        desired unit for keyword, e.g., "m/s"

    Returns
    -------
    Q\_
        Returns a Quantity

    Raises
    ------
    ValueError
        if keywword is neither number, str or quantity

    """

    from esbmtk import Q_

    if isinstance(quantity, str):
        quantity = Q_(quantity)
    elif isinstance(quantity, float | int):
        quantity = Q_(f"{quantity} {unit}")
    elif not isinstance(quantity, Q_):
        raise ValueError("kw must be string, number or Quantity")

    return quantity


def map_units(obj: any, v: any, *args) -> float:
    """parse v to see if it is a string. if yes, map to quantity.
    parse v to see if it is a quantity, if yes, map to model units
    and extract magnitude, assign mangitude to return value
    if not, assign value to return value

    :param obj: connection object
    :param v: input string/number/quantity
    :args: list of model base units
    :returns: number

    :raises ScaleError: if input cannot be mapped to a model unit
    """
    from . import Q_

    m: float = 0
    match: bool = False

    # test if string, map to quantity if yes
    if isinstance(v, str):
        v = Q_(v)

    # test if we find a matching dimension, map if true
    if isinstance(v, Q_):
        for q in args:
            if v.dimensionality == q.dimensionality:
                m = v.to(q).magnitude
                match = True

        if not match:
            raise ScaleError(
                f"Missing base units for the scale in {obj.full_name}, this should not happen"
                f"complain to the program author"
            )

    else:  # no quantity, so it should be a number
        m = v

    if not isinstance(m, (int, float)):
        raise ValueError(f"m is {type(m)}, must be float, v={v}. Something is fishy")

    return m


def __find_flux__(reservoirs: list, full_name: str):
    """Helper function to find a Flux object based on its full_name in the reservoirs
    in the list of provided reservoirs.

    PRECONDITIONS: full_name must contain the full_name of the Flux

    Parameters:
        reservoirs: List containing all reservoirs
        full_name: str specifying the full name of the flux (boxes.flux_name)
    """
    needed_flux = None
    for res in reservoirs:
        for flux in res.lof:
            if flux.full_name == full_name:
                needed_flux = flux
                break
        if needed_flux is not None:
            break
    if needed_flux is None:
        raise NameError(
            f"add_carbonate_system: Flux {full_name} cannot be found in any of the reservoirs in the Model!"
        )

    return needed_flux


def __checktypes__(av: dict[any, any], pv: dict[any, any]) -> None:
    """this method will use the the dict key in the user provided
    key value data (pv) to look up the allowed data type for this key in av

    av = dictinory with the allowed input keys and their type
    pv = dictionary with the user provided key-value data
    """

    k: any
    v: any

    # loop over provided keywords
    for k, v in pv.items():
        # check av if provided value v is of correct type
        if av[k] != any and not isinstance(v, av[k]):
            raise TypeError(
                f"{type(v)} is the wrong type for '{k}', should be '{av[k]}'"
            )


def dict_alternatives(d: dict, e: str, a: str) -> any:
    """
    The =dict_alternatives= function takes a dictionary =d=, an expression =e=,
    and an alternative expression =a=. It returns the value associated with
    either =a= or =e= in the dictionary =d=.

    :param d: A dictionary.
    :param e: The first expression to check.
    :param a: The alternative expression to check.

    :returns r: The value associated with either =a= or =e= in the dictionary =d=.

    :raises ValueError: If neither =a= nor =e= are found in the dictionary.
    """
    if a in d.keys():
        r = d[a]
    elif e in d:
        r = d[e]
    else:
        raise ValueError(f"You must specify either {e} or {a} \nd = {d}")

    return r


def __checkkeys__(lrk: list, lkk: list, kwargs: dict) -> None:
    """check if the mandatory keys are present

    lrk = list of required keywords
    lkk = list of all known keywords
    kwargs = dictionary with key-value pairs

    """

    k: str
    v: any
    # test if the required keywords are given
    for k in lrk:  # loop over required keywords
        if isinstance(k, list):  # If keyword is a list
            s: int = 0  # loop over allowed substitutions
            for e in k:  # test how many matches are in this list
                if (
                    e in kwargs
                    and not isinstance(e, (np.ndarray, list))
                    and kwargs[e] != "None"
                ):
                    s = s + 1
            if s > 1:  # if more than one match
                raise ValueError(f"You need to specify exactly one from this list: {k}")

        elif k not in kwargs:
            raise ValueError(f"You need to specify a value for {k}")

    tl: list[str] = [k for k, v in lkk.items()]
    # test if we know all keys
    for k in kwargs:
        if k not in lkk:
            raise ValueError(f"{k} is not a valid keyword. \n Try any of \n {tl}\n")


def __addmissingdefaults__(lod: dict, kwargs: dict) -> dict:
    """
    test if the keys in lod exist in kwargs, otherwise add them with the default values
    from lod

    """

    new: dict = {}
    if lod:
        for k, v in lod.items():
            if k not in kwargs:
                new[k] = v

    # kwargs |= new
    kwargs.update(new)
    return kwargs


def data_summaries(
    M: Model,
    species_names: list,
    box_names: list,
    register_with="None",
) -> list:
    """Group results by species and ReservoirGroups

    :param M: model instance
    :param species_names: list of species instances
    :param box_names: list of ReservoirGroup instances
    :param register_with: defaults to M
    :returns pl: a list of datafield instance to be plotted

    """

    from esbmtk import DataField, VectorData

    if register_with == "None":
        register_with = M

    pl = []
    for sp in species_names:
        data_list = []
        label_list = []
        t = ""
        for b in box_names:
            if isinstance(b, VectorData):
                data_list.append(a)
                label_list.append(a.full_name)
                t = b.name
            if hasattr(b, f"{sp.name}"):
                a = getattr(b, f"{sp.name}")  # this is a reservoir
                y1, y1_label, unit = a.get_plot_format()
                data_list.append(y1)
                label_list.append(f"{b.name}")
                if hasattr(sp, "display_as"):
                    t = sp.display_as
                else:
                    t = sp.name

        df = DataField(
            name=f"{sp.name}_df",
            register=register_with,
            x1_data=M.time,
            y1_data=data_list,
            y1_label=label_list,
            y1_legend=f"{sp.name} [{unit}]",
            x1_as_time=True,
            title=t,
        )
        pl.append(df)
        """ check if species has isotope data. This needs to be done
        after the above loop, so that the data is in its own datafield.
        """
        data_list = []
        label_list = []
        for b in box_names:
            if hasattr(b, f"{sp.name}"):
                a = getattr(b, f"{sp.name}")
                if a.isotopes:
                    unit = f"{a.species.element.d_label}"
                    data_list.append(a.d)
                    label_list.append(f"{b.name}")

        if len(data_list) > 0:
            df = DataField(
                name=f"d_{sp.name}_df",
                register=register_with,
                x1_data=M.time,
                y1_data=data_list,
                y1_label=label_list,
                y1_legend=f"{unit} [{sp.element.d_scale}]",
                x1_as_time=True,
                title=sp.element.d_label,
            )
            pl.append(df)

    return pl


@njit(fastmath=True)
def get_l_mass(m: float, d: float, r: float) -> float:
    """
    :param m: mass or concentration
    :param d: delta value
    :param r: isotopic reference ratio

    return mass or concentration of the light isotopeb
    """
    return (1000.0 * m) / ((d + 1000.0) * r + 1000.0)


def get_delta_h(R) -> float:
    """Calculate the delta of a flux or reservoir

    :param R: Reservoir or Flux handle

    returns d as vector of delta values
    R.c = total concentration
    R.l = concentration of the light isotope
    """

    from esbmtk import Reservoir, GasReservoir, Flux

    r = R.species.r  # reference ratio
    if isinstance(R, (Reservoir, GasReservoir)):
        d = np.where(R.l > 0, 1e3 * ((R.c - R.l) / R.l - r) / r, 0)
    elif isinstance(R, Flux):
        d = np.where(R.l > 0, 1e3 * ((R.m - R.l) / R.l - r) / r, 0)
    else:
        raise ValueError(
            f"{R.full_name} must be of type Flux or Reservoir, not {type(R)}"
        )

    return d


def get_delta_from_concentration(c, l, r):
    """Calculate the delta from the mass of light and heavy isotope

    :param c: total mass/concentration
    :param l: light isotope mass/concentration
    :param r: reference ratio

    """
    h = c - l
    d = 1000 * (h / l - r) / r
    return d


def get_imass(m: float, d: float, r: float) -> [float, float]:
    """
    Calculate the isotope masses from bulk mass and delta value.
    Arguments are m = mass, d= delta value, r = abundance ratio
    species

    """

    l: float = (1000.0 * m) / ((d + 1000.0) * r + 1000.0)
    h: float = m - l
    return [l, h]


@njit(fastmath=True)
def get_delta(l: NDArrayFloat, h: NDArrayFloat, r: float) -> NDArrayFloat:
    """Calculate the delta from the mass of light and heavy isotope

    :param l: light isotope mass/concentration
    :param h: heavy isotope mass/concentration
    :param r: reference ratio

    :return : delta

    """
    return 1000 * (h / l - r) / r


@njit(fastmath=True)
def get_new_ratio_from_alpha(
    ref_mass: float,  # reference mass
    ref_l: float,  # reference light istope
    a: float,  # fractionation factor
) -> [float, float]:
    """Calculate the effect of the istope fractionation factor alpha on
    the ratio between the mass of the light isotope devided by the total mass

    Note that alpha needs to be given as fractional value, i.e., 1.07 rather
    than 70 (i.e., (alpha-1) * 1000
    """

    if ref_mass > 0.0:
        new_ratio = -ref_l / (a * ref_l - a * ref_mass - ref_l)
    else:
        new_ratio = 0.0

    return new_ratio


def register_user_function(M: Model, lib_name: str, func_name: str | list) -> None:
    """Register user supplied library and function with the model

    Parameters
    ----------
    M : Model
        Model handle
    lib_name : str
        name of python file that contains the function
    func_name : str | list
        Name of one or more user supplied function(s)

    """
    import pathlib as pl
    import sys

    if isinstance(func_name, str):
        func_name = [func_name]

    cwd: pl.Path = pl.Path.cwd()  # get the current working
    if cwd not in sys.path:
        sys.path.append(cwd)

    for f in func_name:
        M.luf[f] = [lib_name, cwd]



# import classes from the esbmtk library
from esbmtk import (
    Model,  # the model class
    Connection,
    ReservoirGroup,  # the reservoir class
    ConnectionGroup,  # the connection class
    SourceGroup,  # the source class
    SinkGroup,  # sink class
    Q_,  # Quantity operator
    data_summaries,
)

# define the basic model parameters
M = Model(
    stop="500 yr",  # end time of model
    timestep="1 yr",  # upper limit of time step
    element=["Phosphor", "Carbon"],  # list of element definitions
)

# Fixed ratios
Redfield = 130  # C:P ratio
R_e = 0.01  # About 1% of the exported P is buried in the deep ocean
tau = Q_("100 year")  # PO4 residence time in surface box
thc = "20*Sv"  # Thermohaline circulation in Sverdrup
F_w_OM = M.set_flux("5850 Gmol", "year", M.C)  # P @280 ppm (Filipelli 2002)
F_w_PO4 = F_w_OM / Redfield

# Source definitions
SourceGroup(
    register=M,  # i.e., the instance will be available as M.weathering
    name="weathering",
    species=[M.DIC, M.PO4],
    isotopes={M.DIC: True, M.PO4: False},
)
SinkGroup(
    register=M,  #
    name="burial",
    species=[M.DIC, M.PO4],
    isotopes={M.DIC: True, M.PO4: False},
)

ReservoirGroup(
    name="S_b",
    register=M,
    volume="3E16 m**3",  # surface box volume
    concentration={M.DIC: "1 umol/l", M.PO4: "0 umol/l"},
    isotopes={M.DIC: True},
    delta={M.DIC: 0},
)

ReservoirGroup(
    name="D_b",  # box name
    register=M,  # this box will be available M.db
    volume="100E16 m**3",  # deeb box volume
    concentration={M.DIC: "1 umol/l", M.PO4: "0 umol/l"},  # species in box
    isotopes={M.DIC: True},
    delta={M.DIC: 0},
)

ConnectionGroup(
    source=M.weathering,  # source of flux
    sink=M.S_b,  # target of flux
    rate={M.DIC: F_w_OM, M.PO4: F_w_PO4},  # rate of flux
    delta={M.DIC: 0},
    ctype="regular",  # required!
    id="weathering",  # connection id
)

ConnectionGroup(  # thermohaline downwelling
    source=M.S_b,  # source of flux
    sink=M.D_b,  # target of flux
    ctype="scale_with_concentration",
    scale=thc,
    id="downwelling",
)
ConnectionGroup(  # thermohaline upwelling
    source=M.D_b,  # source of flux
    sink=M.S_b,  # target of flux
    ctype="scale_with_concentration",
    scale=thc,
    id="upwelling",
)

# Primary production as a function of P-concentration
Connection(  #
    source=M.S_b.DIC,  # source of flux
    sink=M.D_b.DIC,  # target of flux
    ref_reservoirs=M.S_b.PO4,
    ctype="scale_with_concentration",
    scale=Redfield * M.S_b.volume / tau,
    id="OM_production",
    alpha=-28,  # mUr
)

# POP export as a funtion of OM export
Connection(  #
    source=M.S_b.PO4,  # source of flux
    sink=M.D_b.PO4,  # target of flux
    ctype="scale_with_flux",
    ref_flux=M.flux_summary(filter_by="OM_production", return_list=True)[0],
    scale=1 / Redfield,
    id="POP",
)

# P burial
Connection(  #
    source=M.D_b.PO4,  # source of flux
    sink=M.burial.PO4,  # target of flux
    ctype="scale_with_flux",
    ref_flux=M.flux_summary(filter_by="POP", return_list=True)[0],
    scale=R_e,
    id="P_burial",
)

# OM burial
Connection(  #
    source=M.D_b.DIC,  # source of flux
    sink=M.burial.DIC,  # target of flux
    ctype="scale_with_flux",
    ref_flux=M.flux_summary(filter_by="OM_production", return_list=True)[0],
    scale=R_e,
    id="OM_burial",
    # alpha=0,
)


M.run()
pl = data_summaries(
    M,
    [M.DIC],  # species
    [M.S_b, M.D_b],  # boxes
)
M.plot(pl, fn="po4_2_with_isotopes.png")
# M.save_state(directory="state_po42")



def my_burial(concentration: float, p: tuple) -> float:
    """Calculate a flux as a function of concentration

    Parameters
    ----------
    concentration : float
        substance concentration
    p : tuple
        tuple where the first element is the scaling factor

    Returns
    -------
    float
        flux in model mass unit / time

    Notes: the scale information is passed as a tuple, so we need
    extract it from the tuple before using it

    f is a burial flux, so we need to return a negative number.
    """
    (scale,) = p

    f = concentration * scale

    return -f


def add_my_burial(source, sink, species, scale) -> None:
    """This function initializes a user supplied function
    so that it can be used within the ESBMTK eco-system

    Parameters
    ----------
    source : Source | Reservoir | ReservoirGroup
        A source
    sink : Sink | Reservoir | ReservoirGroup
        A sink
    species : Species
        A model species
    scale : float
        A scaling factor

    """
    from esbmtk import ExternalCode, register_return_values

    p = (scale,)  # convert float into tuple
    ec = ExternalCode(
        name="mb",
        species=source.species,
        function=my_burial,
        fname="my_burial",
        ftype="std",
        function_input_data=[source],
        function_params=p,
        register=source,
        return_values=[
            {f"F_{sink.full_name}.{species.name}": "id_string"},
        ],
    )

    register_return_values(ec, source)



"""
esbmtk: A general purpose Earth Science box model toolkit Copyright
(C), 2020 Ulrich G.  Wortmann

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or (at
your option) any later version.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
from __future__ import annotations
from pandas import DataFrame
import time
from time import process_time
import numpy as np
import typing as tp
import numpy.typing as npt
import matplotlib.pyplot as plt
import pandas as pd
import logging
import os
import psutil
import warnings
from . import ureg, Q_

from .esbmtk_base import esbmtkBase

from .utility_functions import (
    show_data,
    plot_geometry,
    find_matching_strings,
    get_l_mass,
    get_delta_h,
    get_delta_from_concentration,
)

# declare numpy types
NDArrayFloat = npt.NDArray[np.float64]

if tp.TYPE_CHECKING:
    from .extended_classes import ExternalData, DataField
    from .connections import Connection
    from .processes import Process


class ModelError(Exception):
    def __init__(self, message):
        message = f"\n\n{message}\n"
        super().__init__(message)


class ReservoirError(Exception):
    def __init__(self, message):
        message = f"\n\n{message}\n"
        super().__init__(message)


class FluxError(Exception):
    def __init__(self, message):
        message = f"\n\n{message}\n"
        super().__init__(message)


class ScaleError(Exception):
    def __init__(self, message):
        message = f"\n\n{message}\n"
        super().__init__(message)


class Model(esbmtkBase):
    """This class is used to specify a new model.  See the __init__()
    method for a detailed explanation of the parameters

    The user facing methods of the model class are

        - Model_Name.info()
        - Model_Name.save_data()
        - Model_Name.plot([sb.DIC, sb.TA]) plot any object in the list
        - Model_Name.save_state() Save the model state
        - Model_name.read_state() Initialize with a previous model
          state
        - Model_Name.run()
        - Model_Name.list_species()
        - Model_name.flux_summary()
        - Model_Name.connection_summary()

    """

    def __init__(self, **kwargs: dict[any, any]) -> None:
        """Initialize a model instance

        :param **kwargs: A dictionary with key value pairs.

            .. Example::

                    esbmtkModel(name   =  "Test_Model", # required
                                stop     = "10000 yrs", # end time
                                timestep = "1 yr",    # as a string "2 yrs"
                                element = ["Carbon", "Sulfur" ]
                              )

                :param name: The model name, e.g., M
                :param mass_unit: mol
                :param volume_unit: only tested with liter
                :param element: list with one or more species names
                :param time_step: Limit automatic step size increase, i.e., the time
                resolution of the model. Optional, defaults to the model duration/100
                :param m_type: enables or disables isotope calculation for the
                entire model.  The default value is "Not set" in this case
                isotopes will only be calculated for reservoirs which set
                the isotope keyword.  'mass_only' 'both' will override the
                reservoir settings

                :param offset: will offset the time axis by the specified
                amount, when plotting the data, .i.e., the model time runs
                from to 100, but you want to plot data as if where from
                2000 to 2100, you would specify a value of 2000.  This is
                for display purposes only, and does not affect the model.
                Care must be taken that any external data references the
                model time domain, and not the display time.

                :param display_precision: affects the on-screen display of data.
                It is also cutoff for the graphical output.  I.e., the
                interval f the y-axis will not be smaller than the
                display_precision.

                :param opt_k_carbonic: see  https://doi.org/10.5194/gmd-15-15-2022

                :param opt_pH_scale: total = 1, free = 3,
        """
        from importlib.metadata import version
        import datetime
        from esbmtk import species_definitions, hypsometry

        self.defaults: dict[str, list[any, tuple]] = {
            "start": ["0 yrs", (str, Q_)],
            "stop": ["None", (str, Q_)],
            "offset": ["0 yrs", (str, Q_)],
            "timestep": ["None", (str, Q_)],
            "element": ["None", (str, list)],
            "mass_unit": ["mol", (str)],
            "volume_unit": ["liter", (str)],
            "area_unit": ["m**2", (str)],
            "time_unit": ["year", (str)],
            "concentration_unit": ["mol/liter", (str)],
            "time_label": ["Years", (str)],
            "display_precision": [0.01, (float)],
            "plot_style": ["default", (str)],
            "m_type": ["Not Set", (str)],
            "number_of_datapoints": [1000, (int)],
            "step_limit": [1e9, (int, float, str)],
            "register": ["local", (str)],
            "save_flux_data": [False, (bool)],
            "full_name": ["None", (str)],
            "parent": ["None", (str)],
            "isotopes": [False, (bool)],
            "debug": [False, (bool)],
            "ideal_water": [True, (bool)],
            "use_ode": [True, (bool)],
            "parse_model": [True, (bool)],
            "rtol": [1.0e-6, (float)],
            "bio_pump_functions": [0, (int)],  # custom/old
            # "area_table": [None, (str, np.ndarray)],
            "opt_k_carbonic": [15, (int)],
            "opt_pH_scale": [1, (int)],  # 1: total scale
            "opt_buffers_mode": [2, (int)],
        }

        # provide a list of absolutely required keywords
        self.lrk: list[str] = [
            "stop",
            "timestep",
        ]
        self.__initialize_keyword_variables__(kwargs)

        self.name = "M"

        # empty list which will hold all reservoir references
        self.lmo: list = []
        self.lmo2: list = []
        self.dmo: dict = {}  # dict of all model objects. useful for name lookups

        # start a log file
        for handler in logging.root.handlers[:]:
            logging.root.removeHandler(handler)

        fn: str = f"{self.name}.log"
        logging.basicConfig(filename=fn, filemode="w", level=logging.CRITICAL)
        self.__register_name_new__()

        self.lor: list = []  # list of all reservoir type objects
        self.lic: list = []  # list of all reservoir type objects
        # empty list which will hold all connector references
        self.loc: set = set()  # set with connection handles
        self.lel: list = []  # list which will hold all element references
        self.led: list[ExternalData] = []  # all external data objects
        self.lsp: list = []  # list which will hold all species references
        self.lop: list = []  # set of flux processes
        self.lpc_f: list = []  # list of external functions affecting fluxes
        # list of external functions that needs to be imported in the ode_backend
        self.lpc_i: list = []
        # list of external functions affecting virtual reservoirs
        self.lpc_r: list = []
        self.lvr: list = [] # list of virtual reservoirs
        # optional keywords for use in the connector class
        self.olkk: list = []
        # list of objects which require a delayed initialize
        self.lto: list = []
        self.ldf: list = [] # list of datafield objects
        self.los: list = []  # list of signals
        self.lof: list = []  # list of fluxes
        self.lrg: list = []  # list of reservoirgroups
        self.gpt: tuple = ()  # global parameter list
        self.toc: tuple = ()  # global constants list
        self.gcc: int = 0  # constants counter
        self.vpc: int = 0  # parameter counter
        self.luf: dict = {}  # user functions and source

        # unit defs
        self.l_unit = ureg.meter  # the length unit
        self.t_unit = Q_(self.time_unit).units  # the time unit
        self.d_unit = Q_(self.stop).units  # display time units
        self.m_unit = Q_(self.mass_unit).units  # the mass unit
        self.v_unit = Q_(self.volume_unit).units  # the volume unit
        self.a_unit = Q_(self.area_unit).units  # the area unit
        self.c_unit = Q_(self.concentration_unit).units
        self.f_unit = self.m_unit / self.t_unit  # the flux unit (mass/time)
        self.r_unit = self.v_unit / self.t_unit  # flux as volume/time

        # legacy variable names
        self.start = self.ensure_q(self.start).to(self.t_unit).magnitude
        self.stop = self.ensure_q(self.stop).to(self.t_unit).magnitude
        self.timestep = self.ensure_q(self.timestep)
        self.dt = self.timestep.to(self.t_unit).magnitude
        self.max_step = self.dt
        self.offset = self.ensure_q(self.offset).to(self.t_unit).magnitude
        self.start = self.start + self.offset
        self.stop = self.stop + self.offset
        self.n = self.name
        self.mo = self.name
        self.model = self
        self.plot_style: list = [self.plot_style]

        self.xl = f"Time [{self.t_unit}]"  # time axis label
        self.length = int(abs(self.stop - self.start))
        self.steps = int(abs(round(self.length / self.dt)))

        self.time = (np.arange(self.steps) * self.dt) + self.start
        self.time_ode = np.linspace(
            self.start,
            self.stop - (self.stop - self.start) / 100,
            num=100,
        )
        self.timec = np.empty(0)
        self.state = 0

        # calculate stride
        self.stride = int(self.steps / self.number_of_datapoints)
        self.reset_stride = self.stride

        if self.step_limit == "None":
            self.number_of_solving_iterations: int = 0
        elif self.step_limit > self.steps:
            self.number_of_solving_iterations: int = 0
            self.step_limit = "None"
        else:
            self.step_limit = int(self.step_limit)
            self.number_of_solving_iterations = int(round(self.steps / self.step_limit))
            self.reset_stride = int(round(self.steps / self.number_of_datapoints))
            self.steps = self.step_limit
            self.time = (np.arange(self.steps) * self.dt) + self.start

        # set_printoptions(precision=self.display_precision)

        if "element" in self.kwargs:
            if isinstance(self.kwargs["element"], list):
                element_list = self.kwargs["element"]
            else:
                element_list = [self.kwargs["element"]]

            # register elements and species with model
            for e in element_list:
                # get function handle
                fh = getattr(species_definitions, e)
                fh(self)  # register element with model
                # get element handle
                eh = getattr(self, e)
                # register species with model
                eh.__register_species_with_model__()

        warranty = (
            f"\n"
            f"ESBMTK {version('esbmtk')}  Copyright (C) 2020 - {datetime.date.today().year}  Ulrich G.Wortmann\n"
            f"This program comes with ABSOLUTELY NO WARRANTY\n"
            f"For details see the LICENSE file\n"
            f"This is free software, and you are welcome to redistribute it\n"
            f"under certain conditions; See the LICENSE file for details.\n"
        )
        print(warranty)

        # initialize the hypsometry class
        hypsometry(name="hyp", model=self, register=self)

    def info(self, **kwargs) -> None:
        """Show an overview of the object properties.  Optional
        arguments are (name/default/explanation)

        :param index: int = 0 # this will show data at the given index
        :param indent: int = 0 # print indentation
        """
        off: str = "  "

        if "indent" not in kwargs:
            indent = 0
            ind = ""
        else:
            indent = kwargs["indent"]
            ind = " " * indent

        # print basic data bout this object
        print(self)

        # list elements
        print("Currently defined elements and their species:")
        for e in self.lel:
            print(f"{ind}{e}")
            print(f"{off} Defined Species:")
            for s in e.lsp:
                print(f"{off}{off}{ind}{s.n}")

    def save_state(self, directory="state") -> None:
        """Save model state.  Similar to save data, but only saves the
        last 10 time-steps
        """
        from pathlib import Path
        from esbmtk import rmtree

        fn: str = directory  # file name
        cwd: Path = Path.cwd()  # get the current working directory
        fqfn: Path = Path(f"{cwd}/{fn}")  # fully qualified file name
        if fqfn.exists():  # check if file exist
            print(f"Found previous state, deleting {fqfn}")
            rmtree(fqfn)
            if fqfn.exists():
                raise FileExistsError(f"Deleting {fqfn} failed")

        start: int = -2
        stop: int = None
        stride: int = 1
        prefix: str = "state_"

        for r in self.lor:  # loop over reservoirs
            r.__write_data__(prefix, start, stop, stride, False, directory)

    def save_data(self, directory="./data") -> None:
        """Save the model results to a CSV file. Each reservoir will have
        their own CSV file

        Calling save_data() without any arguments,  will create (or
        recreate) the data directory in the current working directory
        which will then be populated by csv-files

        :param directory: a string with the directory name. It defaults to 'data'

        """
        from pathlib import Path
        from esbmtk import rmtree

        fn: str = directory  # file name
        cwd: Path = Path.cwd()  # get the current working directory
        fqfn: Path = Path(f"{cwd}/{fn}")  # fully qualified file name
        if fqfn.exists():  # check if file exist
            print(f"Found previous state, deleting {fqfn}")
            rmtree(fqfn)
            if fqfn.exists():
                raise FileExistsError(f"Deleting {fqfn} failed")

        prefix = ""
        stride = self.stride
        start = 0
        stop = self.steps
        append = False

        for r in self.lor:
            if r.rtype != "flux_only":
                r.__write_data__(prefix, start, stop, stride, append, directory)

    def read_data(self, directory="./data") -> None:
        """Save the model results to a CSV file. Each reservoir will have
        their own CSV file
        """
        from esbmtk import Reservoir, GasReservoir

        prefix = ""

        print(f"reading data from {directory}")

        for r in self.lor:
            if isinstance(r, (Reservoir, GasReservoir)):
                r.__read_state__(directory, prefix)
                if r.isotopes:
                    r.d = get_delta_from_concentration(r.c, r.l, r.sp.r)

    def restart(self):
        """Restart the model with result of the last run.  This is
        useful for long runs which otherwise would used to much memory
        """

        for r in self.lor:
            r.__reset_state__()
            for f in r.lof:
                f.__reset_state__()

        for r in self.lvr:
            r.__reset_state__()

        # print(f"len of time {len(self.time)}, stride = {self.stride}")
        # print(f"len of time with stride {len(self.time[0 : -2 : self.stride])}")
        self.timec = np.append(self.timec, self.time[0 : -2 : self.stride])
        t = int(round((self.stop - self.start) * self.dt))
        self.start = int(round((self.stop * self.dt)))
        self.stop = self.start + t
        self.time = (np.arange(self.steps) * self.dt) + self.start
        print(f"new start = {self.start}, new stop = {self.stop}")
        print(f"time[0] = {self.time[0]} time[-1] = {self.time[-1]}")
        # print(f"len of timec {len(self.timec)}")
        # self.time = (arange(self.steps) * self.dt) + self.start

    def read_state(self, directory="state"):
        """This will initialize the model with the result of a
        previous model run.  For this to work, you will need issue a
        Model.save_state() command at then end of a model run.  This
        will create the necessary data files to initialize a
        subsequent model run.
        """

        from esbmtk import Reservoir, GasReservoir  # GasReservoir

        for r in self.lor:
            if isinstance(r, (Reservoir, GasReservoir)):
                r.__read_state__(directory)

        # update swc object
        for rg in self.lrg:
            if hasattr(rg, "swc"):
                rg.swc.update_parameters(pos=0)

    def merge_temp_results(self):
        """Replace the datafields which were used for an individual
        iteration with the data we saved from the previous iterations
        """

        self.time = self.timec
        for r in self.lor:
            r.__merge_temp_results__()
            for f in r.lof:
                f.__merge_temp_results__()

        for r in self.lvr:
            r.__merge_temp_results__()

    def plot(self, pl: list = None, **kwargs) -> None:
        """Plot all objects specified in pl

        :param pl: a list of ESBMTK instance (e.g., reservoirs)

        optional keywords: fn = filename, defaults to the Model name

        Example::

            M.plot([sb.PO4, sb.DIC], fn='test.pdf')

        will plot sb.PO4 and sb.DIC and save the plot as 'test.pdf'
        """

        if pl is None:
            pl = []
        filename = kwargs.get("fn", f"{self.n}.pdf")
        noo: int = len(pl)
        size, [row, col] = plot_geometry(noo)  # adjust layout
        fig, ax = plt.subplots(row, col)  # row, col
        axs = [[], []]

        """ The shape of the ax value of subplots depends on the figure
        geometry. So we need to ensure we are dealing with a 2-D array
        """
        if row == 1 and col == 1:  # row=1, col=1 only one window
            axs = ax
        elif row > 1 and col == 1:  # mutiple rows, one column
            axs = []
            for i in range(row):
                axs.append(ax[i])
        elif row == 1 and col:  # 1 row, multiple columns
            print(f"one row, mutiple cols")
            for i in range(geo[1]):
                axs.append(ax[i])
        else:
            axs = ax  # mutiple rows and mutiple columns

        # ste plot parameters
        plt.style.use(self.plot_style)
        fig.canvas.manager.set_window_title(f"{self.n} Reservoirs")
        fig.set_size_inches(size)

        i = 0  # loop over objects
        for c in range(row):  # rows
            if col > 1:
                for r in range(col):  # columns
                    if i < noo:
                        pl[i].__plot__(self, axs[c][r])
                        i = i + 1
                    else:
                        axs[c][r].remove()
            elif row > 1:
                pl[i].__plot__(self, axs[c])
                i = i + 1
            else:
                pl[i].__plot__(self, axs)
                i = i + 1

        fig.subplots_adjust(top=0.88)
        fig.tight_layout()
        plt.show(block=False)  # create the plot windows
        fig.savefig(filename)

    def run(self, **kwargs) -> None:
        """Loop over the time vector, and for each time step, calculate the
        fluxes for each reservoir
        """

        # this has nothing todo with self.time below!
        wts = time.time()
        start: float = process_time()
        # new: NDArrayFloat = np.zeros(4)

        solver = "ode" if "solver" not in kwargs else kwargs["solver"]
        self.solver = solver
        self.__run_solver__(solver, kwargs)

        self.state = 1  # flag that the model has executed
        duration: float = process_time() - start
        wcd = time.time() - wts
        print(f"\n Execution took {duration:.2f} cpu seconds, wt = {wcd:.2f}\n")
        process = psutil.Process(os.getpid())
        print(f"This run used {process.memory_info().rss/1e9:.2f} Gbytes of memory \n")

    def get_delta_values(self):
        """Calculate masses and isotope ratios in the usual delta notation"""
        for r in self.lor:
            if r.isotopes:
                r.m = r.c * r.volume
                r.d = get_delta_h(r)

    def sub_sample_data(self):
        """Subsample the data.  No need to save 100k lines of data You
        need to do this _after_ saving the state, but before plotting
        and saving the data
        """

        stride = int(len(self.time) / self.number_of_datapoints)

        if stride > 1:
            self.time = self.time[2:-2:stride]

            for r in self.lor:
                r.__sub_sample_data__(stride)

            for vr in self.lvr:
                vr.__sub_sample_data__(stride)

            for f in self.lof:
                f.__sub_sample_data__(stride)

    def __run_solver__(self, solver: str, kwargs: dict) -> None:
        if solver == "ode":
            self.ode_solver(kwargs)
        else:
            raise ModelError(f"Solver={self.solver} is unkknown")

    def ode_solver(self, kwargs):
        """
        Use the ode solver
        """
        from esbmtk import Q_, write_equations_2, get_initial_conditions
        from scipy.integrate import solve_ivp
        import sys
        import pathlib as pl

        # build equation file
        R, icl, cpl, ipl, atol = get_initial_conditions(self, self.rtol)
        self.R = R
        self.icl = icl
        self.cpl = cpl
        self.ipl = ipl

        # write equation system
        if self.parse_model:
            eqs_file = write_equations_2(self, R, icl, cpl, ipl)
            # print(f"loc = {eqs_file.resolve()}")
        else:
            warnings.warn("Reusing equation file\n")

        # ensure that cwd is in the load path. Required for windows
        cwd: pl.Path = pl.Path.cwd()
        sys.path.append(cwd)
      
        from equations import eqs   # import equation system
       
        method = kwargs["method"] if "method" in kwargs else "BDF"
        stype = kwargs["stype"] if "stype" in kwargs else "solve_ivp"

        # check if using carbonate chemistry, if not provide dummy values
        if not hasattr(self, "area_table"):
            self.area_table = 0
            self.area_dz_table = 0
            self.Csat_table = 0

        if stype == "solve_ivp":
            self.results = solve_ivp(
                eqs,
                (self.time[0], self.time[-1]),
                R,
                args=(
                    self,
                    self.gpt,
                    self.toc,
                    self.area_table,
                    self.area_dz_table,
                    self.Csat_table,
                ),
                method=method,
                atol=atol,
                rtol=self.rtol,
                t_eval=self.time_ode,
                first_step=Q_("1 second").to(self.t_unit).magnitude,
                max_step=self.max_step,
            )

        print(
            f"\nnfev={self.results.nfev}, njev={self.results.njev}, nlu={self.results.nlu}\n"
        )
        print(f"status={self.results.status}")
        print(f"message={self.results.message}\n")

        self.post_process_data(self.results)

    def test_d_pH(self, rg: ReservoirGroup, time: NDArrayFloat) -> None:
        """Test if the change in pH exceeds more than 0.01 units per time step.
        Note that this is only a crude test, since the solver interpolates
        between intergration steps. So this may not catch all problems.

        Parameters
        ----------
        rg : ReservoirGroup
            ReservoirGroup instance
        time: : NDArrayFloat
            time vector as returned by the solver
        """
        hplus = getattr(rg, "Hplus")
        dph = np.diff(-np.log10(hplus.c))
        dph_bool = dph > 0.01
        if sum(dph_bool) > 0:
            for i, v in enumerate(dph_bool):
                if v:
                    warnings.warn(
                        f"\n\n{rg.full_name} delta pH = {dph[i]:.2f} "
                        f"at t = {time[i]:.2f} {self.t_unit:~P}\n"
                    )

    def post_process_data(self, results) -> None:
        """Map solver results back into esbmtk structures

        :param results: numpy arrays with solver results
        """

        from esbmtk import carbonate_system_1_pp

        # interpolate signals into the ode time domain
        # must be done before changing model time domain
        for s in self.los:
            s.data.m = np.interp(
                results.t,
                self.time,
                s.data.m,
            )
            if s.isotopes:
                s.data.l = np.interp(
                    results.t,
                    self.time,
                    s.data.l,
                )

        # interpolate external data into ode time domain
        # must be done before changing model time domain
        for ed in self.led:
            ed.y = np.interp(results.t, ed.x, ed.y)
        # assign results to the esbmtk variables
        i = 0
        for r in self.icl:
            r.c = results.y[i]
            # this needs fixing if we have variable volumes
            r.m = results.y[i] * r.volume
            i = i + 1
            if r.isotopes:
                r.l = results.y[i]
                i = i + 1
                r.d = get_delta_from_concentration(r.c, r.l, r.sp.r)

        self.time = results.t
        steps = len(results.t)  # get number of solver steps
        for f in self.lof:
            if f.save_flux_data:
                f.m = f.m[0:steps]
                if f.isotopes:
                    f.l = f.l[0:steps]
                    f.d = get_delta_h(f)

        # test for rapid changes in pH and
        # calculate carbonate species for cs1
        for rg in self.lrg:
            if hasattr(rg, "Hplus"):
                self.test_d_pH(rg, results.t)
            if rg.has_cs1:
                carbonate_system_1_pp(rg)

    def list_species(self):
        """List all  defined species."""
        for e in self.lel:
            print(f"{e.n}")
            e.list_species()

    def flux_summary(self, **kwargs: dict):
        """Show a summary of all model fluxes

        Optional parameters:

        :param filter_by: str = "" # filter on flux name or part of
            flux name words separated by blanks act as additional
            conditions, i.e., all words must occur in a given name

        :param return_list: bool = False, # if True return a list of
            fluxes matching the filter_by string.

        :param exclude: str = "" # exclude all results matching this
            string

            Example::

                names = M.flux_summary(filter_by="POP A_sb", return_list=True)
        """

        fby = ""
        rl: list = []
        exclude: str = ""
        check_exlusion: bool = False

        if "exclude" in kwargs:
            exclude = kwargs["exclude"]
            check_exlusion = True

        if "filter_by" in kwargs:
            fby: list = kwargs["filter_by"].split(" ")

        if "filter" in kwargs:
            raise ModelError("use filter_by instead of filter")

        if "return_list" in kwargs:
            return_list = True
        else:
            return_list = False
            print(f"\n --- Flux Summary -- filtered by {fby}\n")

        for f in self.lof:  # loop over flux list
            if find_matching_strings(f.full_name, fby) and (
                check_exlusion and exclude not in f.full_name or not check_exlusion
            ):
                rl.append(f)
                if not return_list:
                    print(f"{f.full_name}")
        if not return_list:
            rl = None

        return rl

    def connection_summary(self, **kwargs: dict) -> None:
        """Show a summary of all connections

        Optional parameters:

        :param filter_by: str = "" # filter on connection id. If more than one word is provided, all words must match

        :param return_list: bool if set, return a list object instead of printing to the terminal

        """

        rl = []
        if "filter_by" in kwargs:
            fby: list = kwargs["filter_by"].split(" ")
        else:
            fby: bool = False

        if "filter" in kwargs:
            raise ModelError("use filter_by instead of filter")

        if "silent" not in kwargs:
            print(f"\n --- Connection Group Summary -- filtered by {fby}\n")
            print(f"       run the following command to see more details:\n")

        # test if all words of the fby list occur in c.full_name. If yes,
        self.cg_list: list = list(list(self.loc))
        for c in self.cg_list:
            if fby and find_matching_strings(c.full_name, fby) or not fby:
                if "silent" in kwargs or "return_list" in kwargs:
                    rl.append(c)
                else:
                    print(f"{c.full_name}.info()\n")
        return rl

    def clear(self):
        """
        delete all model objects
        """
        for o in self.lmo:
            print(f"deleting {o}")
            del __builtins__[o]


class Element(esbmtkBase):
    """Each model, can have one or more elements.  This class sets
    element specific properties

    Example::

        Element(name      = "S "           # the element name
                model     = Test_model     # the model handle
                mass_unit =  "mol",        # base mass unit
                li_label  =  "$^{32$S",    # Label of light isotope
                hi_label  =  "$^{34}S",    # Label of heavy isotope
                d_label   =  r"$\delta^{34}$S",  # Label for delta value
                d_scale   =  "VCDT",       # Isotope scale
                r         = 0.044162589,   # isotopic abundance ratio for element
              )
    """

    # set element properties
    def __init__(self, **kwargs) -> any:
        """Initialize all instance variables Defaults are as follows::

            self.defaults: dict[str, list[any, tuple]] = {
               "name": ["M", (str)],
               "model": ["None", (str, Model)],
               "register": ["None", (str, Model)],
               "full_name": ["None", (str)],
               "li_label": ["None", (str)],
               "hi_label": ["None", (str)],
               "d_label": ["None", (str)],
               "d_scale": ["None", (str)],
               "r": [1, (float, int)],
               "mass_unit": ["mol", (str, Q_)],
               "parent": ["None", (str, Model)],

        }

        Required keywords: "name", "model", "mass_unit"
        """
        self.defaults: dict[str, list[any, tuple]] = {
            "name": ["M", (str)],
            "model": ["None", (str, Model)],
            "register": ["None", (str, Model)],
            "full_name": ["None", (str)],
            "li_label": ["None", (str)],
            "hi_label": ["None", (str)],
            "d_label": ["None", (str)],
            "d_scale": ["None", (str)],
            "r": [1, (float, int)],
            "mass_unit": ["", (str, Q_)],
            "parent": ["None", (str, Model)],
        }

        # list of absolutely required keywords
        self.lrk: list = ["name", "model", "mass_unit"]
        self.__initialize_keyword_variables__(kwargs)

        self.parent = self.model
        # legacy name aliases
        self.n: str = self.name  # display name of species
        self.mo: Model = self.model  # model handle
        self.mu: str = self.mass_unit  # display name of mass unit
        self.ln: str = self.li_label  # display name of light isotope
        self.hn: str = self.hi_label  # display name of heavy isotope
        self.dn: str = self.d_label  # display string for delta
        self.ds: str = self.d_scale  # display string for delta scale
        self.lsp: list = []  # list of species for this element.
        self.mo.lel.append(self)

        if self.mo.register == "local" and self.register == "None":
            self.register = self.mo

        self.__register_name_new__()

    def list_species(self) -> None:
        """
        List all species which are predefined for this element
        """

        for e in self.lsp:
            print(e.n)

    def __register_species_with_model__(self) -> None:
        """
        Bit of hack, but makes model code more readable
        """

        for s in self.lsp:
            setattr(self.model, s.name, s)


class Species(esbmtkBase):
    """Each model, can have one or more species.  This class sets
    species specific properties

    Example::

        Species(name = "SO4",
                element = S,

    )

    Defaults::

        self.defaults: dict[any, any] = {
            "name": ["None", (str)],
            "element": ["None", (Element, str)],
            "display_as": [kwargs["name"], (str)],
            "m_weight": [0, (int, float, str)],
            "register": ["None", (Model, Element, Reservoir, GasReservoir)],
            "parent": ["None", (Model, Element, Reservoir, GasReservoir)],
            "flux_only": [False, (bool)],
            "logdata": [False, (bool)],
            "scale_to": ["None", (str)],
            "stype": ["concentration", (str)],
        }

    Required keywords: "name", "element"
    """

    # set species properties
    def __init__(self, **kwargs) -> None:
        """Initialize all instance variables"""

        from esbmtk import GasReservoir

        # provide a list of all known keywords
        self.defaults: dict[any, any] = {
            "name": ["None", (str)],
            "element": ["None", (Element, str)],
            "display_as": [kwargs["name"], (str)],
            "m_weight": [0, (int, float, str)],
            "register": [kwargs["element"], (Model, Element, Reservoir, GasReservoir)],
            "parent": ["None", (Model, Element, Reservoir, GasReservoir)],
            "flux_only": [False, (bool)],
            "logdata": [False, (bool)],
            "scale_to": ["None", (str)],
            "stype": ["concentration", (str)],
        }

        # provide a list of absolutely required keywords
        self.lrk = ["name", "element"]
        self.__initialize_keyword_variables__(kwargs)
        self.parent = self.register

        if "display_as" not in kwargs:
            self.display_as = self.name

        # legacy names
        self.n = self.name  # display name of species
        self.mass_unit = self.element.mass_unit
        self.mu = self.mass_unit  # display name of mass unit
        self.ln = self.element.ln  # display name of light isotope
        self.hn = self.element.hn  # display name of heavy isotope
        self.dn = self.element.dn  # display string for delta
        self.ds = self.element.ds  # display string for delta scale
        self.r = self.element.r  # ratio of isotope standard
        self.mo = self.element.mo  # model handle
        self.eh = self.element.n  # element name
        self.e = self.element  # element handle
        self.dsa = self.display_as  # the display string.
        # self.mo.lsp.append(self)   # register self on the list of model objects
        self.e.lsp.append(self)  # register this species with the element

        if self.mo.register == "local" and self.register == "None":
            self.register = self.mo
        self.__register_name_new__()


class ReservoirBase(esbmtkBase):
    """
    Base class for all Reservoir objects
    """

    def __init__(self, **kwargs) -> None:
        raise NotImplementedError(
            "ReservoirBase should never be used. Use the derived classes"
        )

    def __set_legacy_names__(self, kwargs) -> None:
        """
        Move the below out of the way
        """

        from esbmtk import get_box_geometry_parameters

        self.atol: list[float] = [1.0, 1.0]  # tolerances
        self.lof: list[Flux] = []  # flux references
        self.led: list[ExternalData] = []  # all external data references
        self.lio: dict[str, int] = {}  # flux name:direction pairs
        self.lop: list[Process] = []  # list holding all processe references
        self.loe: list[Element] = []  # list of elements in thiis reservoir
        self.doe: dict[Species, Flux] = {}  # species flux pairs
        self.loc: set[Connection] = set()  # set of connection objects
        self.ldf: list[DataField] = []  # list of datafield objects
        # list of processes which calculate reservoirs
        self.lpc: list[Process] = []
        self.ef_results = False  # Reservoir has external function results

        # legacy names
        self.n: str = self.name  # name of reservoir
        # if "register" in self.kwargs:
        if self.register == "None":
            self.pt = self.name
        else:
            self.pt: str = f"{self.register.name}_{self.n}"
            self.groupname = self.register.name
            # self.full_name = f"{self.register.name}.{self.n}"
        # else:
        #   self.pt = self.name

        self.sp: Species = self.species  # species handle
        self.mo: Model = self.species.mo  # model handle
        self.model = self.mo
        self.rvalue = self.sp.r
        self.m_unit = self.model.m_unit
        self.v_unit = self.model.v_unit
        self.c_unit = self.model.c_unit

        # right y-axis label
        self.ld: str = f"{self.species.dn} [{self.species.ds}]"
        self.xl: str = self.mo.xl  # set x-axis lable to model time

        if self.legend_left == "None":
            self.legend_left = self.species.dsa

        self.legend_right = f"{self.species.dn} [{self.species.ds}]"
        # legend_left is in __init__ !

        # decide whether we use isotopes
        if self.mo.m_type == "both":
            self.isotopes = True
        elif self.mo.m_type == "mass_only":
            self.isotopes = False

        if self.geometry != "None" and self.geometry_unset:
            get_box_geometry_parameters(self)

        if self.display_precision == 0:
            self.display_precision = self.mo.display_precision

        self.parent = self.register

    # setup a placeholder setitem function
    def __setitem__(self, i: int, value: float):
        return self.__set_data__(i, value)

    def __call__(self) -> None:  # what to do when called as a function ()
        return self

    def __getitem__(self, i: int) -> NDArrayFloat:
        """
        Get flux data by index
        """

        return np.array([self.m[i], self.l[i], self.c[i]])

    def __set_with_isotopes__(self, i: int, value: float) -> None:
        """
        :param i: index
        :param value: array of [mass, li, hi, d]

        """
        self.m[i]: float = value[0]
        # update concentration and delta next. This is computationally inefficient
        # but the next time step may depend on on both variables.
        self.c[i]: float = value[0] / self.v[i]  # update concentration
        self.l[i]: float = value[1] / self.v[i]  # update concentration

    def __set_without_isotopes__(self, i: int, value: float) -> None:
        """
        :param i: index
        :param value: array of [mass]

        """
        self.m[i]: float = value[0]
        self.c[i]: float = self.m[i] / self.v[i]  # update concentration

    def __update_mass__() -> None:
        """
        Place holder function
        """

        raise NotImplementedError("__update_mass__ is not yet implmented")

    def __write_data__(
        self,
        prefix: str,
        start: int,
        stop: int,
        stride: int,
        append: bool,
        directory: str,
    ) -> None:
        """Write data to file.  This function is called by the
        write_data() and save_state() methods

        :param prefix:
        :param start:
        :param stop:
        :param stride:
        :param append:
        :param directory:
        """
        from pathlib import Path

        p = Path(directory)
        p.mkdir(parents=True, exist_ok=True)

        # some short hands
        sn = self.sp.n  # species name
        sp = self.sp  # species handle
        mo = self.sp.mo  # model handle

        smu = f"{mo.m_unit:~P}"
        mtu = f"{mo.t_unit:~P}"
        fmu = f"{mo.f_unit:~P}"
        cmu = f"{mo.c_unit:~P}"

        # sdn = self.sp.dn  # delta name
        # sds = self.sp.ds  # delta scale
        rn = self.full_name  # reservoir name
        mn = self.sp.mo.n  # model name
        if self.sp.mo.register == "None":
            fn = f"{directory}/{prefix}{mn}_{rn}.csv"  # file name
        elif self.sp.mo.register == "local":
            fn = f"{directory}/{prefix}{rn}.csv"  # file name
        else:
            raise ReservoirError(
                f"Model register keyword must be 'None'/'local' not {self.sp.mo.register}"
            )

        # build the dataframe
        df: pd.dataframe = DataFrame()

        # breakpoint()
        df[f"{rn} Time [{mtu}]"] = self.mo.time[start:stop:stride]  # time
        # df[f"{rn} {sn} [{smu}]"] = self.m.to(self.mo.m_unit).magnitude[start:stop:stride]  # mass
        if self.isotopes:
            # print(f"rn = {rn}, sp = {sp.name}")
            df[f"{rn} {sp.ln} [{cmu}]"] = self.l[start:stop:stride]  # light isotope
        df[f"{rn} {sn} [{cmu}]"] = self.c[start:stop:stride]  # concentration

        file_path = Path(fn)
        if append and file_path.exists():
            df.to_csv(file_path, header=False, mode="a", index=False)
        else:
            df.to_csv(file_path, header=True, mode="w", index=False)
        return df

    def __sub_sample_data__(self, stride) -> None:
        """There is usually no need to keep more than a thousand data
        points so we subsample the results before saving, or
        processing them
        """

        # print(f"Reset data with {len(self.m)}, stride = {self.mo.reset_stride}")
        self.m = self.m[2:-2:stride]
        self.l = self.l[2:-2:stride]
        self.c = self.c[2:-2:stride]

    def __reset_state__(self) -> None:
        """Copy the result of the last computation back to the
        beginning so that a new run will start with these values

        save the current results into the temp fields
        """

        # print(f"Reset data with {len(self.m)}, stride = {self.mo.reset_stride}")
        self.mc = np.append(self.mc, self.m[0 : -2 : self.mo.reset_stride])
        # self.dc = np.append(self.dc, self.d[0 : -2 : self.mo.reset_stride])
        self.cc = np.append(self.cc, self.c[0 : -2 : self.mo.reset_stride])

        # copy last result into first field
        self.m[0] = self.m[-2]
        self.l[0] = self.l[-2]
        # self.h[0] = self.h[-2]
        # self.d[0] = self.d[-2]
        self.c[0] = self.c[-2]

    def __merge_temp_results__(self) -> None:
        """Once all iterations are done, replace the data fields with
        the saved values
        """

        self.m = self.mc
        self.c = self.cc
        # self.d = self.dc

    def __read_state__(self, directory: str, prefix="state_") -> None:
        """read data from csv-file into a dataframe
        The CSV file must have the following columns
            - Model Time t
            - Reservoir_Name m
            - Reservoir_Name l
            - Reservoir_Name h
            - Reservoir_Name d
            - Reservoir_Name c
            - Flux_name m
            - Flux_name l etc etc.
        """

        from .utility_functions import is_name_in_list, get_object_from_list

        read: set = set()
        curr: set = set()

        if self.sp.mo.register == "None":
            fn = f"{directory}/{prefix}{self.mo.n}_{self.full_name}.csv"
        elif self.sp.mo.register == "local":
            fn = f"{directory}/{prefix}{self.full_name}.csv"
        else:
            raise ReservoirError(
                f"Model register keyword must be 'None'/'local' not {self.sp.mo.register}"
            )

        if not os.path.exists(fn):
            raise FileNotFoundError(
                f"Flux {fn} does not exist in Reservoir {self.full_name}"
            )

        self.df: pd.DataFrame = pd.read_csv(fn)
        self.headers: list = list(self.df.columns.values)
        df = self.df
        headers = self.headers

        # the headers contain the object name for each data in the
        # reservoir or flux thus, we must reduce the list to unique
        # object names first. Note, we must preserve order
        header_list: list = []
        for x in headers:
            n = x.split(" ")[0]
            if n not in header_list:
                header_list.append(n)

        # loop over all columns
        col: int = 1  # we ignore the time column
        i: int = 0
        for n in header_list:
            name = n.split(" ")[0]
            logging.debug(f"Looking for {name}")
            # this finds the reservoir name
            if name == self.full_name:
                # logging.debug(f"found reservoir data for {name}")
                col = self.__assign_reservoir_data__(self, df, col, True)
            else:
                raise ReservoirError(f"Unable to find Flux {n} in {self.full_name}")

        # test if we missed any fluxes
        for f in list(curr.difference(read)):
            warnings.warn(f"\nDid not find values for {f}\n in saved state")

    def __assign_reservoir_data__(
        self, obj: any, df: pd.DataFrame, col: int, res: bool
    ) -> int:
        """Assign the third last entry data to all values in reservoir

        :param obj: # Reservoir
        :param df: pd.dataframe
        :param col: int # index into column position
        :param res: True # indicates whether obj is reservoir

        :returns: int # index into last column
        """

        # this may need fixing
        if obj.isotopes:
            obj.l[:] = df.iloc[-1, col]  # get last row
            col += 1
            obj.c[:] = df.iloc[-1, col]
            col += 1
        else:
            v = df.iloc[-1, col]
            obj.c[:] = df.iloc[-1, col]
            col += 1

        return col

    def get_plot_format(self):
        """Return concentrat data in plot units"""
        from pint import Unit

        if isinstance(self.plt_units, Q_):
            unit = f"{self.plt_units.units:~P}"
        elif isinstance(self.plt_units, Unit):
            unit = f"{self.plt_units:~P}"
        else:
            unit = f"{self.plt_units}"

        y1_label = f"{self.legend_left} [{unit}]"

        if self.display_as == "mass":
            y1 = (self.m * self.mo.m_unit).to(self.plt_units).magnitude
        elif self.display_as == "ppm":
            y1 = self.c * 1e6
            y1_label = "ppm"
        elif self.display_as == "length":
            y1 = (self.c * self.mo.l_unit).to(self.plt_units).magnitude
        else:
            y1 = (self.c * self.mo.c_unit).to(self.plt_units).magnitude

        # test for plt_transform
        if self.plot_transform_c != "None":
            if callable(self.plot_transform_c):
                y1 = self.plot_transform_c(self.c)
            else:
                raise ReservoirError("Plot transform must be a function")

        return y1, y1_label, unit

    def __plot__(self, M: Model, ax) -> None:
        """Plot Model data

        :param M: Model
        :param ax: # graph axes handle
        """

        from esbmtk import set_y_limits

        # convert time and data to display units
        x = (M.time * M.t_unit).to(M.d_unit).magnitude

        y1, y1_label, unit = self.get_plot_format()

        # plot first axis
        ax.plot(x[1:-2], y1[1:-2], color="C0", label=y1_label)
        ax.set_xlabel(f"{M.time_label} [{M.d_unit:~P}]")
        ax.set_ylabel(y1_label)

        # add any external data if present
        for i, d in enumerate(self.led):
            leg = f"{self.lm} {d.legend}"
            ax.scatter(d.x[1:-2], d.y[1:-2], color=f"C{i+2}", label=leg)

        ax.spines["top"].set_visible(False)
        handler1, label1 = ax.get_legend_handles_labels()
        set_y_limits(ax, self)

        if self.isotopes:
            axt = ax.twinx()
            y2 = self.d  # no conversion for isotopes
            axt.plot(x[1:-2], y2[1:-2], color="C1", label=self.legend_right)
            axt.set_ylabel(self.ld)
            set_y_limits(axt, self)
            ax.spines["top"].set_visible(False)
            # set combined legend
            handler2, label2 = axt.get_legend_handles_labels()
            axt.legend(handler1 + handler2, label1 + label2, loc=0).set_zorder(6)
        else:
            ax.legend(handler1, label1)
            ax.spines["right"].set_visible(False)
            ax.yaxis.set_ticks_position("left")
            ax.xaxis.set_ticks_position("bottom")

        ax.set_title(self.full_name)

    def info(self, **kwargs) -> None:
        """Show an overview of the object properties.  Optional
        arguments are

        :param index: int = 0 # this will show data at the given index
        :param indent: int = 0 # print indentation
        """

        off: str = "  "
        index = 0 if "index" not in kwargs else kwargs["index"]
        if "indent" not in kwargs:
            indent = 0
            ind = ""
        else:
            indent = kwargs["indent"]
            ind = " " * indent

        # print basic data bout this reservoir
        print(f"{ind}{self.__str__(kwargs)}")
        print(f"{ind}Data sample:")
        show_data(self, index=index, indent=indent)

        print(f"\n{ind}Connnections:")
        for p in sorted(self.loc):
            print(f"{off}{ind}{p.full_name}.info()")

        print(f"\n{ind}Fluxes:")

        # m = Q_("1 Sv").to("l/a").magnitude
        for i, f in enumerate(self.lof):
            print(f"{off}{ind}{f.full_name}: {self.lodir[i]*f.m[-2]:.2e}")

        print()
        print("Use the info method on any of the above connections")
        print("to see information on fluxes and processes")


class Reservoir(ReservoirBase):
    """This object holds reservoir specific information.

    Example::

        Reservoir(name = "foo",      # Name of reservoir
                  species = S,          # Species handle
                  delta = 20,           # initial delta - optional (defaults  to 0)
                  mass/concentration = "1 unit"  # species concentration or mass
                  volume/geometry = "1E5 l",      # reservoir volume (m^3)
                  plot = "yes"/"no", defaults to yes
                  plot_transform_c = a function reference, optional (see below)
                  legend_left = str, optional, useful for plot transform
                  display_precision = number, optional, inherited from Model
                  register = Model instance
                  isotopes = True/False otherwise use Model.m_type
                  seawater_parameters= dict, optional
                  )

    You must either give mass or concentration.  The result will
    always be displayed as concentration though.

    You must provide either the volume or the geometry keyword.  In
    the latter case provide a list where the first entry is the upper
    depth datum, the second entry is the lower depth datum, and the
    third entry is the total ocean area.  E.g., to specify the upper
    200 meters of the entire ocean, you would write:

    geometry=[0,-200,3.6e14]

    the corresponding ocean volume will then be calculated by the
    calc_volume method in this case the following instance variables
    will also be set:

    self.volume in model units (usually liter) self.are:a surface area
    in m^2 at the upper bounding surface self.sed_area: area of
    seafloor which is intercepted by this box.  self.area_fraction:
    area of seafloor which is intercepted by this relative to the
    total ocean floor area

    It is also possible to specify volume and area explicitly. In this
    case provide a dictionary like this::

        geometry = {"area": "1e14 m**2", # surface area
                    "volume": "3e16 m**3", # box volume
                   }

    Adding seawater_properties:
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    If this optional parameter is specified, a SeaWaterConstants instance will
    be registered for this Reservoir as Reservoir.swc See the
    SeaWaterConstants class for details how to specify the parameters,
    e.g.:

    .. code-block:: python

            seawater_parameters = {"temperature": 2,
                                   "pressure": 240,
                                   "salinity" : 35,
                                  }

    Using a transform function:
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    In some cases, it is useful to transform the reservoir
    concentration data before plotting it.  A good example is the H+
    concentration in water which is better displayed as pH.  We can do
    this by specifying a function to convert the reservoir
    concentration into pH units::

    .. code-block:: python

        def phc(c :float) -> float:
            # Calculate concentration as pH. c can be a number or numpy array
            import numpy as np
            pH :float = -np.log10(c)
            return pH

    this function can then be added to a reservoir as:

    hplus.plot_transform_c = phc

    You can modify the left legend to suit the transform via the
    legend_left keyword

    Note, at present the plot_transform_c function will only take one
    argument, which always defaults to the reservoir concentration.
    The function must return a single argument which will be
    interpreted as the transformed reservoir concentration.

    Accesing Reservoir Data:
    ~~~~~~~~~~~~~~~~~~~~~~~~

    You can access the reservoir data as:

        - Name.m # mass

        - Name.d # delta

        - Name.c # concentration

    Useful methods include:

        - Name.write_data() # save data to file

        - Name.info() # info Reservoir

    """

    def __init__(self, **kwargs) -> None:
        """Initialize a reservoir

        Defaults::

            self.defaults: dict[str, list[any, tuple]] = {
              "name": ["None", (str)],
              "species": ["None", (str, Species)],
              "delta": ["None", (int, float, str)],
              "concentration": ["None", (str, Q_, float)],
              "mass": ["None", (str, Q_)],
              "volume": ["None", (str, Q_)],
              "geometry": ["None", (list, dict, str)],
              "plot_transform_c": ["None", (any)],
              "legend_left": ["None", (str)],
              "plot": ["yes", (str)],
              "groupname": ["None", (str)],
              "rtype": ["regular", (str)],
              "function": ["None", (str, col.Callable)],
              "display_precision": [0.01, (int, float)],
              "register": [
                  "None",
                  (SourceGroup, SinkGroup, ReservoirGroup, ConnectionGroup, Model, str),
              ],
              "parent": [
                  "None",
                  (SourceGroup, SinkGroup, ReservoirGroup, ConnectionGroup, Model, str),
              ],
              "full_name": ["None", (str)],
              "seawater_parameters": ["None", (dict, str)],
              "isotopes": [False, (bool)],
              "ideal_water": ["None", (str, bool)],
              "has_cs1": [False, (bool)],
              "has_cs2": [False, (bool)],

        }

        Required Keywords::

            self.lrk: list = [
              "name",
              "species",
              "register",
              ["volume", "geometry"],
              ["mass", "concentration"],

        ]
        """

        from esbmtk import (
            SourceGroup,
            SinkGroup,
            ReservoirGroup,
            ConnectionGroup,
            SeawaterConstants,
            get_box_geometry_parameters,
            phc,
        )

        # provide a dict of all known keywords and their type
        self.defaults: dict[str, list[any, tuple]] = {
            "name": ["None", (str)],
            "species": ["None", (str, Species)],
            "delta": ["None", (int, float, str)],
            "concentration": ["None", (str, Q_, float)],
            "mass": ["None", (str, Q_)],
            "volume": ["None", (str, Q_)],
            "geometry": ["None", (list, dict, str)],
            "geometry_unset": [True, (bool)],
            "plot_transform_c": ["None", (any)],
            "legend_left": ["None", (str)],
            "plot": ["yes", (str)],
            "groupname": ["None", (str)],
            "rtype": ["regular", (str)],
            "function": ["None", (str, callable)],
            "display_precision": [0.01, (int, float)],
            "register": [
                "None",
                (SourceGroup, SinkGroup, ReservoirGroup, ConnectionGroup, Model, str),
            ],
            "parent": [
                "None",
                (SourceGroup, SinkGroup, ReservoirGroup, ConnectionGroup, Model, str),
            ],
            "full_name": ["None", (str)],
            "seawater_parameters": ["None", (dict, str)],
            "isotopes": [False, (bool)],
            "ideal_water": ["None", (str, bool)],
            "has_cs1": [False, (bool)],
            "has_cs2": [False, (bool)],
        }

        # provide a list of absolutely required keywords
        self.lrk: list = [
            "name",
            "species",
            ["volume", "geometry"],
            ["mass", "concentration"],
        ]

        self.__initialize_keyword_variables__(kwargs)

        if isinstance(self.register, Model):
            self.model = self.register
        else:
            self.model = self.register.model
        self.parent = self.register
        self.c = np.zeros(len(self.model.time))
        self.l = np.zeros(len(self.model.time))
        self.m = np.zeros(len(self.model.time))
        self.__set_legacy_names__(kwargs)

        # geoemtry information
        if self.volume == "None":
            get_box_geometry_parameters(self)
        else:
            self.volume = Q_(self.volume).to(self.mo.v_unit)

        self.model.toc = (*self.model.toc, self.volume.to(self.model.v_unit).magnitude)
        self.v_index = self.model.gcc
        self.model.gcc = self.model.gcc + 1
        self.c_unit = self.model.c_unit
        # This should probably be species specific?
        self.mu: str = self.sp.e.mass_unit  # massunit xxxx

        if self.sp.stype == "concentration":
            if self.mass == "None":
                if isinstance(self.concentration, (str, Q_)):
                    cc = Q_(self.concentration)
                    # concentration can be mol/kg or mol/l
                    sm, sc = str(cc.units).split(" / ")  # get
                    mm, mc = str(self.mo.c_unit).split(" / ")  # model
                    if mc == "liter" and sc == "kilogram":
                        cc = Q_(f"{cc.magnitude} {str(self.mo.c_unit)}")
                        warnings.warn(
                            "\nConvert mol/kg to mol/liter assuming density = 1\n"
                        )
                    elif sc != mc:
                        raise ScaleError(
                            f"no transformation for {cc.units} to {self.mo.c_unit}"
                        )
                    self._concentration = cc.to(self.mo.c_unit)
                    self.plt_units = self.mo.c_unit
                else:
                    cc = self.concentration
                    self.plt_units = self.mo.c_unit
                    self._concentration = cc

                self.mass = (
                    self.concentration.to(self.mo.c_unit).magnitude
                    * self.volume.to(self.mo.v_unit).magnitude
                )
                self.mass = Q_(f"{self.mass} {self.mo.c_unit}")
                self.display_as = "concentration"

                # fixme: c should be dimensionless, not sure why this happens
                self.c = self.c.to(self.mo.c_unit).magnitude

                if self.species.scale_to != "None":
                    c, m = str(self.mo.c_unit).split(" / ")
                    self.plt_units = Q_(f"{self.species.scale_to} / {m}")
            elif self.concentration == "None":
                m = Q_(self.mass)
                self.plt_units = self.mo.m_unit
                self.mass: tp.Union[int, float] = m.to(self.mo.m_unit).magnitude
                self.concentration = self.massto(self.mo.m_unit) / self.volume.to(
                    self.mo.v_unit
                )
                self.display_as = "mass"
            else:
                raise ReservoirError("You need to specify mass or concentration")

        elif self.sp.stype == "length":
            self.plt_units = self.mo.l_unit
            self.c = np.zeros(self.mo.steps) + Q_(self.concentration).magnitude
            self.display_as = "length"
        self.state = 0

        # save the unit which was provided by the user for display purposes
        # left y-axis label
        self.lm: str = f"{self.species.n} [{self.mu}/l]"

        # initialize mass vector
        if self.mass == "None":
            self.m: NDArrayFloat = np.zeros(self.mo.steps)
        else:
            self.m: NDArrayFloat = np.zeros(self.species.mo.steps) + self.mass
        self.l: NDArrayFloat = np.zeros(self.mo.steps)
        # self.c: NDArrayFloat = np.zeros(self.mo.steps)
        self.v: NDArrayFloat = (
            np.zeros(self.mo.steps) + self.volume.to(self.mo.v_unit).magnitude
        )  # reservoir volume

        if self.delta != "None":
            self.l = get_l_mass(self.c, self.delta, self.species.r)

        # create temporary memory if we use multiple solver iterations
        if self.mo.number_of_solving_iterations > 0:
            self.mc = np.empty(0)
            self.cc = np.empty(0)
            self.dc = np.empty(0)

        self.mo.lor.append(self)  # add this reservoir to the model
        if self.rtype != "flux_only":
            self.mo.lic.append(self)  # reservoir type object list

        # register instance name in global name space
        if self.mo.register == "local" and self.register == "None":
            self.register = self.mo
        self.__register_name_new__()

        # decide which setitem functions to use
        if self.isotopes:
            self.__set_data__ = self.__set_with_isotopes__
        else:
            self.__set_data__ = self.__set_without_isotopes__

        if self.species.name == "Hplus":
            self.plot_transform_c = phc
        # any auxilliary init - normally empty, but we use it here to extend the
        # reservoir class in virtual reservoirs
        self.__aux_inits__()

    @property
    def concentration(self) -> float:
        return self._concentration

    @property
    def delta(self) -> float:
        return self._delta

    @property
    def mass(self) -> float:
        return self._mass

    # @property
    # def volume(self) -> float:
    #     return self._volume

    # @volume.setter
    # def volume(self) -> None:
    #     self.volume = self._volume.to(self.register.v_unit)

    @concentration.setter
    def concentration(self, c) -> None:
        if self.update and c != "None":
            breakpoint()
            # this requires unit screening
            # then conversion into model units and mganitude
            # followed by updates to c and m
            self._concentration = c.to(self.mo.c_unit)
            self.mass = (
                self._concentration * self.volume * self.density / 1000
            )  # caculate mass
            self.c = self.c * 0 + self._concentration.magnitude
            self.m = self.m * 0 + self.mass

    @delta.setter
    def delta(self, d: float) -> None:
        if self.update and d != "None":
            self._delta: float = d
            self.isotopes = True
            self.l = get_l_mass(self.c, d, self.species.r)

    @mass.setter
    def mass(self, m: float) -> None:
        if self.update and m != "None":
            self._mass: float = m
            """ problem: m_unit can be mole, but data can be in liter * mole /kg
            this should not happen and results in an error converting to magnitide
            """
            self.m = np.zeros(self.species.mo.steps) + m
            self.c = self.m / self.volume.to(self.mo.v_unit).magnitude


class Flux(esbmtkBase):
    """A class which defines a flux object.  Flux objects contain
    information which links them to an species, describe things like
    the mass and time unit, and store data of the total flux rate at
    any given time step.  Similarly, they store the flux of the light
    and heavy isotope flux, as well as the delta of the flux.  This is
    typically handled through the Connect object.  If you set it up
    manually

    Example::

        Flux = (name = "Name" # optional, defaults to _F
             species = species_handle,
             delta = any number,
             rate  = "12 mol/s" # must be a string
             display_precision = number, optional, inherited from Model

    )

    You can access the flux data as

        - Name.m # mass
        - Name.d # delta
        - Name.c # same as Name.m since flux has no concentration
    """

    def __init__(self, **kwargs: dict[str, any]) -> None:
        """
        Initialize a flux.  Arguments are the species name the flux
        rate (mol/year), the delta value and unit

        Defaults::

            self.defaults: dict[str, list[any, tuple]] = {
              "name": ["None", (str)],
              "species": ["None", (str, Species)],
              "delta": [0, (str, int, float)],
              "rate": ["None", (str, Q_, int, float)],
              "plot": ["yes", (str)],
              "display_precision": [0.01, (int, float)],
              "isotopes": [False, (bool)],
              "register": [
                  "None",
                  (
                      str,
                      Reservoir,
                      GasReservoir,
                      Connection,
                      Connect,
                      Signal,
                  ),
              ],
              "save_flux_data": [False, (bool)],
              "id": ["None", (str)],
              "ftype": ["None", (str)],

        }

        Required Keywords: "species", "rate", "register"
        """

        from esbmtk import (
            Q_,
            Reservoir,
            GasReservoir,
            Connect,
            Connection,
            Signal,
        )

        # provide a dict of all known keywords and their type
        self.defaults: dict[str, list[any, tuple]] = {
            "name": ["None", (str)],
            "species": ["None", (str, Species)],
            "delta": [0, (str, int, float)],
            "rate": ["None", (str, Q_, int, float)],
            "plot": ["yes", (str)],
            "display_precision": [0.01, (int, float)],
            "isotopes": [False, (bool)],
            "register": [
                "None",
                (
                    str,
                    Reservoir,
                    GasReservoir,
                    Connection,
                    Connect,
                    Signal,
                ),
            ],
            "save_flux_data": [False, (bool)],
            "id": ["None", (str)],
            "ftype": ["None", (str)],
        }

        # provide a list of absolutely required keywords
        self.lrk: list = ["species", "rate", "register"]
        self.__initialize_keyword_variables__(kwargs)
        self.parent = self.register
        # if save_flux_data is unsepcified, use model default
        # if self.save_flux_data == "None":
        #     self.save_flux_data = self.species.mo.save_flux_data

        # legacy names
        self.n: str = self.name  # name of flux
        self.sp: Species = self.species  # species name
        self.mo: Model = self.species.mo  # model name
        self.model: Model = self.species.mo  # model handle
        self.rvalue = self.sp.r

        if self.display_precision == 0:
            self.display_precision = self.mo.display_precision

        # model units
        self.plt_units = Q_(self.rate).units
        self.mu: str = f"{self.species.mu}/{self.mo.t_unit}"

        # and convert flux into model units
        if isinstance(self.rate, str):
            self.rate: float = Q_(self.rate).to(self.mo.f_unit).magnitude
        elif isinstance(self.rate, Q_):
            self.rate: float = self.rate.to(self.mo.f_unit).magnitude
        elif isinstance(self.rate, (int, float)):
            self.rate: float = self.rate

        li = get_l_mass(self.rate, self.delta, self.sp.r) if self.delta else 0
        self.fa: NDArrayFloat = np.asarray([self.rate, li])

        # in case we want to keep the flux data
        if self.save_flux_data:
            self.m: NDArrayFloat = (
                np.zeros(self.model.steps) + self.rate
            )  # add the flux

            if self.isotopes:
                self.l: NDArrayFloat = np.zeros(self.model.steps)
                if self.rate != 0:
                    self.l = get_l_mass(self.m, self.delta, self.species.r)
                    self.fa[1] = self.l[0]

            if self.mo.number_of_solving_iterations > 0:
                self.mc = np.empty(0)
                self.dc = np.empty(0)

        else:
            # setup dummy variables to keep existing numba data structures
            self.m = np.zeros(2)
            self.l = np.zeros(2)

            if self.rate != 0:
                self.fa[1] = get_l_mass(self.fa[0], self.delta, self.species.r)

        self.lm: str = f"{self.species.n} [{self.mu}]"  # left y-axis a label
        self.ld: str = f"{self.species.dn} [{self.species.ds}]"  # right y-axis a label

        self.legend_left: str = self.species.dsa
        self.legend_right: str = f"{self.species.dn} [{self.species.ds}]"

        self.xl: str = self.model.xl  # se x-axis label equal to model time
        self.lop: list[Process] = []  # list of processes
        self.lpc: list = []  # list of external functions
        self.led: list[ExternalData] = []  # list of ext data
        self.source: str = ""  # Name of reservoir which acts as flux source
        self.sink: str = ""  # Name of reservoir which acts as flux sink

        if self.name == "None":
            if isinstance(self.parent, (Connection, Connect)):
                self.name = "_F"
                self.n = self.name
            else:
                self.name = f"{self.id}_F"

        self.__register_name_new__()
        self.mo.lof.append(self)  # register with model flux list

        # decide which setitem functions to use
        # decide whether we use isotopes
        if self.mo.m_type == "both":
            self.isotopes = True
        elif self.mo.m_type == "mass_only":
            self.isotopes = False

        if self.isotopes:
            self.__set_data__ = self.__set_with_isotopes__
            # self.__get_data__ = self.__get_with_isotopes__
        else:
            self.__set_data__ = self.__set_without_isotopes__
            # self.__get_data__ = self.__get_without_isotopes__

    # setup a placeholder setitem function
    def __setitem__(self, i: int, value: NDArrayFloat):
        return self.__set_data__(i, value)

    def __getitem__(self, i: int) -> NDArrayFloat:
        """
        Get data by index
        """
        # return self.__get_data__(i)
        return self.fa

    def __set_with_isotopes__(self, i: int, value: NDArrayFloat) -> None:
        """
        Write data by index
        """

        self.m[i] = value[0]
        self.l[i] = value[1]
        self.fa = value[:4]

    def __set_without_isotopes__(self, i: int, value: NDArrayFloat) -> None:
        """
        Write data by index
        """

        self.fa = [value[0], 0]
        self.m[i] = value[0]

    def __call__(self) -> None:  # what to do when called as a function ()
        pass
        return

    def __add__(self, other):
        """
        adding two fluxes works for the masses, but not for delta
        """

        self.fa = self.fa + other.fa
        self.m = self.m + other.m
        self.l = self.l + other.l

    def __sub__(self, other):
        """
        substracting two fluxes works for the masses, but not for
        delta
        """

        self.fa = self.fa - other.fa
        self.m = self.m - other.m
        self.l = self.l - other.l

    def info(self, **kwargs) -> None:
        """Show an overview of the object properties.  Optional
        arguments are:

        :param index: int = 0 this will show data at the given index
        :param indent: int = 0 indentation
        """
        index = 0 if "index" not in kwargs else kwargs["index"]
        if "indent" not in kwargs:
            indent = 0
            ind = ""
        else:
            indent = kwargs["indent"]
            ind = " " * indent

        # print basic data bout this object
        print(f"{ind}{self.__str__(kwargs)}")
        print(f"{ind}Data sample:")
        show_data(self, index=index, indent=indent)

        if len(self.lop) > 0:
            self._extracted_from_info_27(ind)
        else:
            print("There are no processes for this flux")

    # TODO Rename this here and in `info`
    def _extracted_from_info_27(self, ind):
        print(f"\n{ind}Process(es) acting on this flux:")
        off: str = "  "
        for p in self.lop:
            print(f"{off}{ind}{p.__repr__()}")

        print("")
        print(
            "Use help on the process name to get an explanation what this process does"
        )
        if self.register == "None":
            print(f"e.g., help({self.lop[0].n})")
        else:
            print(f"e.g., help({self.register.name}.{self.lop[0].name})")

    def __plot__(self, M: Model, ax) -> None:
        """Plot instructions.

        :param M: Model
        :param ax: matplotlib axes handle
        """

        from esbmtk import set_y_limits

        # convert time and data to display units
        x = (M.time * M.t_unit).to(M.d_unit).magnitude
        y1 = (self.m * M.m_unit).to(self.plt_units).magnitude

        # test for plt_transform
        if self.plot_transform_c != "None":
            if callable(self.plot_transform_c):
                y1 = self.plot_transform_c(self.c)
            else:
                raise FluxError("Plot transform must be a function")

        # plot first axis
        ax.plot(x[1:-2], y1[1:-2], color="C0", label=self.legend_left)
        ax.set_xlabel(f"{M.time_label} [{M.d_unit:~P}]")
        ax.set_ylabel(self.legend_left)
        ax.spines["top"].set_visible(False)
        handler1, label1 = ax.get_legend_handles_labels()

        # plot second axis
        if self.isotopes:
            axt = ax.twinx()
            y2 = self.d  # no conversion for isotopes
            ln2 = axt.plot(x[1:-2], y2[1:-2], color="C1", label=self.legend_right)
            axt.set_ylabel(self.data.ld)
            set_y_limits(axt, M)
            ax.spines["top"].set_visible(False)
            # set combined legend
            handler2, label2 = axt.get_legend_handles_labels()
            legend = axt.legend(handler1 + handler2, label1 + label2, loc=0).set_zorder(
                6
            )
        else:
            ax.legend(handler1, label1)
            ax.spines["right"].set_visible(False)
            ax.yaxis.set_ticks_position("left")
            ax.xaxis.set_ticks_position("bottom")
        ax.set_title(self.full_name)

    def __sub_sample_data__(self, stride) -> None:
        """There is usually no need to keep more than a thousand data
        points so we subsample the results before saving, or
        processing them
        """

        if self.save_flux_data:
            self.m = self.m[2:-2:stride]
            self.l = self.m[2:-2:stride]

    def __reset_state__(self) -> None:
        """Copy the result of the last computation back to the
        beginning so that a new run will start with these values.

        Also, copy current results into temp field
        """

        if self.save_flux_data:
            self.mc = np.append(self.mc, self.m[0 : -2 : self.mo.reset_stride])
            # copy last element to first position
            self.m[0] = self.m[-2]
            self.l[0] = self.l[-2]

    def __merge_temp_results__(self) -> None:
        """Once all iterations are done, replace the data fields with
        the saved values
        """

        self.m = self.mc


class SourceSink(esbmtkBase):
    """
    This is a meta class to setup a Source/Sink objects.  These are
    not actual reservoirs, but we stil need to have them as objects
    Example::

        Sink(name = "Pyrite",
            species = SO4,
            display_precision = number, optional, inherited from Model
            delta = number or str. optional defaults to "None"
            register = Model handle
        )
    """

    def __init__(self, **kwargs) -> None:
        """
        Defaults::

            self.defaults: dict[str, list[any, tuple]] = {
               "name": ["None", (str)],
               "species": ["None", (str, Species)],
               "display_precision": [0.01, (int, float)],
               "register": [
                   "None",
                   (
                       SourceGroup,
                       SinkGroup,
                       ReservoirGroup,
                       ConnectionGroup,
                       Model,
                       str,
                   ),
               ],
               "delta": ["None", (str, int, float)],
               "isotopes": [False, (bool)],

        Required Keywords: "name", "species"
        """

        from esbmtk import (
            SourceGroup,
            SinkGroup,
            ReservoirGroup,
            ConnectionGroup,
        )

        self.defaults: dict[str, list[any, tuple]] = {
            "name": ["None", (str)],
            "species": ["None", (str, Species)],
            "display_precision": [0.01, (int, float)],
            "register": [
                "None",
                (
                    SourceGroup,
                    SinkGroup,
                    ReservoirGroup,
                    ConnectionGroup,
                    Model,
                    str,
                ),
            ],
            "delta": ["None", (str, int, float)],
            "isotopes": [False, (bool)],
        }
        # provide a list of absolutely required keywords
        self.lrk: list[str] = ["name", "species"]
        self.__initialize_keyword_variables__(kwargs)

        if self.register == "None":  # use a sensible default
            self.register = self.species.element.register

        self.loc: set[Connection] = set()  # set of connection objects

        # legacy names
        # if self.register != "None":
        #    self.full_name = f"{self.name}.{self.register.name}"
        self.parent = self.register
        self.n = self.name
        self.sp = self.species
        self.mo = self.species.mo
        self.model = self.species.mo
        self.u = self.species.mu + "/" + str(self.species.mo.t_unit)
        self.lio: list = []
        self.m = 1  # set default mass and concentration values
        self.c = 1
        self.mo.lic.append(self)  # add source to list of res type objects

        if self.mo.register == "local" and self.register == "None":
            self.register = self.mo
        elif self.register == "None":
            self.pt = self.name
        else:
            self.pt: str = f"{self.register.name}_{self.n}"
            self.groupname = self.register.name

        if self.display_precision == 0:
            self.display_precision = self.mo.display_precision

        self.__register_name_new__()
        self.mo.lic.remove(self)

    @property
    def delta(self):
        return self._delta

    @delta.setter
    def delta(self, d):
        """Set/Update delta"""
        if d != "None":
            self._delta = d
            self.isotopes = True
            self.m = 1
            self.c = 1
            self.l = get_l_mass(self.c, d, self.species.r)
            # self.c = self.l / (self.m - self.l)
            # self.provided_kwargs.update({"delta": d})


class Sink(SourceSink):
    """
    This is a meta class to setup a Source/Sink objects.  These are
    not actual reservoirs, but we stil need to have them as objects
    Example::

        Sink(name = "Pyrite",
            species = SO4,
            display_precision = number, optional, inherited from Model
            delta = number or str. optional defaults to "None"
            register = Model handle
        )
    """


class Source(SourceSink):
    """
    This is a meta class to setup a Source/Sink objects.  These are
    not actual reservoirs, but we stil need to have them as objects
    Example::

        Ssource(name = "weathering",
            species = SO4,
            display_precision = number, optional, inherited from Model
            delta = number or str. optional defaults to "None"
            register = Model handle
        )
    """



"""esbmtk: A general purpose Earth Science box model toolkit Copyright
     (C), 2020-2021 Ulrich G. Wortmann

     This program is free software: you can redistribute it and/or
     modify it under the terms of the GNU General Public License as
     published by the Free Software Foundation, either version 3 of
     the License, or (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see
     <https://www.gnu.org/licenses/>.

"""
from __future__ import annotations
import typing as tp
import numpy as np
from numba import njit
import numpy.typing as npt
from math import log, sqrt
from esbmtk import ExternalCode
from esbmtk.utility_functions import (
    __checkkeys__,
    __addmissingdefaults__,
    __checktypes__,
    register_return_values,
    check_for_quantity,
)

if tp.TYPE_CHECKING:
    from .esbmtk import SeawaterConstants, ReservoirGroup, Flux, Q_

# declare numpy types
NDArrayFloat = npt.NDArray[np.float64]

"""
Carbonate System 1 setup requires 3 steps: First we define the actual function,
carbonate_system_1_ode().  In the second step we create a wrapper
init_carbonate_system_1() that defines how to integrate this function into
esbmtk.  In the third step we create a functiom that uses
init_carbonate_system_1() to associates cs1 instances with the respective
reservoirs.

The process for cs2 is analogous
"""


# @njit(fastmath=True)
def get_hplus(dic, ta, h0, boron, K1, K2, KW, KB) -> float:
    """Calculate H+ concentration based on a previous estimate
    [H+]. After Follows et al. 2006,
    doi:10.1016/j.ocemod.2005.05.004

    :param dic: DIC in mol/kg
    :param ta: TA in mol/kg
    :param h0: initial guess for H+ mol/kg
    :param boron: boron concentration
    :param K1: Ksp1
    :param K2: Ksp2
    :param KW: K_water
    :param KB: K_boron

    :returns H: new H+ concentration in mol/kg
    """
    oh = KW / h0
    boh4 = boron * KB / (h0 + KB)
    fg = h0 - boh4 - oh
    cag = ta + fg
    gamm = dic / cag
    dummy = (1 - gamm) ** 2 * K1**2 - 4.0 * K1 * K2 * (1.0 - 2.0 * gamm)

    return 0.5 * ((gamm - 1.0) * K1 + sqrt(dummy))


# @njit(fastmath=True)
def carbonate_system_1(dic, ta, hplus_0, co2aq_0, p) -> tuple:
    """Calculates and returns the H+ and carbonate alkalinity concentrations
     for the given reservoirgroup

    :param dic: float with the dic concentration
    :param ta: float with the ta concentration
    :param hplus_0: float with the H+ concentration
    :param co2aq_0: float with the [CO2]aq concentration
    :param p: tuple with the parameter list
    :returns:  dCdt_Hplus, dCdt_co2aq

    LIMITATIONS:
    - Assumes all concentrations are in mol/kg
    - Assumes your Model is in mol/kg ! Otherwise, DIC and TA updating will not
    be correct.

    Calculations are based off equations from:
    Boudreau et al., 2010, https://doi.org/10.1029/2009GB003654
    Follows, 2006, doi:10.1016/j.ocemod.2005.05.004
    """

    k1, k2, k1k2, KW, KB, boron, isotopes = p
    if isotopes:  # dic = (x1, x2)
        dic = dic[0]

    hplus = get_hplus(dic, ta, hplus_0, boron, k1, k2, KW, KB)
    co2aq = dic / (1 + k1 / hplus + k1k2 / hplus**2)
    dCdt_Hplus = hplus - hplus_0
    dCdt_co2aq = co2aq - co2aq_0

    return dCdt_Hplus, dCdt_co2aq


def init_carbonate_system_1(rg: ReservoirGroup):
    """Creates a new carbonate system virtual reservoir for each
    reservoir in rgs. Note that rgs must be a list of reservoir groups.

    Required keywords:
        rgs: list = []  of Reservoir Group objects

    These new virtual reservoirs are registered to their respective Reservoir
    as 'cs'.

    The respective data fields are available as rgs.r.cs.xxx where xxx stands
    for a given key key in the  vr_datafields dictionary (i.e., H, CA, etc.)

    """
    from esbmtk import ExternalCode

    p = (
        rg.swc.K1,
        rg.swc.K2,
        rg.swc.K1K2,
        rg.swc.KW,
        rg.swc.KB,
        rg.swc.boron,
        rg.DIC.isotopes,
    )
    ec = ExternalCode(
        name="cs",
        species=rg.mo.Carbon.CO2,
        function=carbonate_system_1,
        fname="carbonate_system_1",
        ftype="std",
        function_input_data=[rg.DIC, rg.TA, "Hplus", "CO2aq"],
        function_params=p,
        register=rg,
        return_values=[
            {f"R_{rg.full_name}.Hplus": rg.swc.hplus},
            {f"R_{rg.full_name}.CO2aq": rg.swc.co2aq},
        ],
    )
    # rg.mo.lpc_f.append(ec.fname)
    rg.mo.lpc_f.append(ec.fname)  # list of function to be imported in ode backend

    return ec


def add_carbonate_system_1(rgs: list):
    """Creates a new carbonate system virtual reservoir for each
    reservoir in rgs. Note that rgs must be a list of reservoir groups.

    Required keywords:
        rgs: list = []  of Reservoir Group objects

    These new virtual reservoirs are registered to their respective Reservoir
    as 'cs'.

    The respective data fields are available as rgs.r.cs.xxx where xxx stands
    for a given key key in the  vr_datafields dictionary (i.e., H, CA, etc.)

    """
    from esbmtk import init_carbonate_system_1

    for rg in rgs:
        if hasattr(rg, "DIC") and hasattr(rg, "TA"):
            ec = init_carbonate_system_1(rg)
            register_return_values(ec, rg)
            rg.has_cs1 = True
        else:
            raise AttributeError(f"{rg.full_name} must have a TA and DIC reservoir")


# @njit(fastmath=True)
def carbonate_system_2(
    CaCO3_export: float,  # 3 CaCO3 export flux as DIC
    dic_t_db: float | tuple,  # 4 DIC in the deep box
    ta_db: float,  # 5 TA in the deep box
    dic_t_sb: float | tuple,  # 6 [DIC] in the surface box
    hplus_0: float,  # 8 hplus in the deep box at t-1
    zsnow: float,  # 9 snowline in meters below sealevel at t-1
    p,
) -> tuple:
    """Calculates and returns the fraction of the carbonate rain that is
    dissolved an returned back into the ocean. This functions returns:

    DIC_burial, DIC_burial_l, Hplus, zsnow

    LIMITATIONS:
    - Assumes all concentrations are in mol/kg
    - Assumes your Model is in mol/kg

    Calculations are based off equations from:
    Boudreau et al., 2010, https://doi.org/10.1029/2009GB003654

    """
    sp, cp, area_table, area_dz_table, Csat_table = p
    ksp0, kc, AD, zsat0, I_caco3, alpha, zsat_min, zmax, z0 = cp
    k1, k2, k1k2, KW, KB, ca2, boron, isotopes = sp

    if isotopes:
        dic_db, dic_db_l = dic_t_db
        dic_sb, dic_sb_l = dic_t_sb
    else:
        dic_db = dic_t_db
        dic_sb = dic_t_sb

    hplus = get_hplus(dic_db, ta_db, hplus_0, boron, k1, k2, KW, KB)
    co3 = max(dic_db / (1 + hplus / k2 + hplus**2 / k1k2), 3.7e-05)

    # ---------- compute critical depth intervals eq after  Boudreau (2010)
    # all depths will be positive to facilitate the use of lookup_tables
    zsat = int(zsat0 * log(ca2 * co3 / ksp0))
    zsat = min(zmax, max(zsat_min, zsat))
    zcc = int(
        zsat0 * log(CaCO3_export * ca2 / (ksp0 * AD * kc) + ca2 * co3 / ksp0)
    )  # eq3
    zcc = min(zmax, max(zsat_min, zcc))
    # get fractional areas
    B_AD = CaCO3_export / AD
    A_z0_zsat = area_table[z0] - area_table[zsat]
    A_zsat_zcc = area_table[zsat] - area_table[zcc]
    A_zcc_zmax = area_table[zcc] - area_table[zmax]
    # ------------------------Calculate Burial Fluxes----------------------------- #
    BCC = A_zcc_zmax * B_AD
    BNS = alpha * A_z0_zsat * B_AD
    diff_co3 = Csat_table[zsat:zcc] - co3
    area_p = area_dz_table[zsat:zcc]
    BDS_under = kc * area_p.dot(diff_co3)
    BDS_resp = alpha * (A_zsat_zcc * B_AD - BDS_under)
    BDS = BDS_under + BDS_resp
    if zsnow > zmax:
        zsnow = zmax
    # integrate saturation difference over area
    diff: NDArrayFloat = Csat_table[zcc : int(zsnow)] - co3
    area_p: NDArrayFloat = area_dz_table[zcc : int(zsnow)]
    BPDC = kc * area_p.dot(diff)
    BPDC = max(BPDC, 0)  # prevent negative values

    """ CACO3_export is the flux of CaCO3 into the box.
    Boudreau's orginal approach is as follows.
    CACO3_export = B_diss + Fburial
    However, the model should use the bypass option and leave all flux
    calculations to the carbonate_system code. As such, ignore the burial flux
    (since it was never added), and only add the fraction of the input flux
    that dissolves back into the box """

    # calculate the differentials
    F_diss = BDS + BCC + BNS + BPDC
    dCdt_Hplus = hplus - hplus_0
    dzdt_zsnow = -BPDC / (area_dz_table[int(zsnow)] * I_caco3)

    """ The isotope ratio of the dissolution flux is determined by the delta
    value of the sediments we are dissolving, and the delta of the carbonate rain.
    The currrent code, assumes that both are the same.
    """
    if isotopes:
        F_diss_l = F_diss * dic_sb_l / dic_sb
        rv = (F_diss, F_diss_l, F_diss * 2, dCdt_Hplus, dzdt_zsnow)
    else:
        rv = (F_diss, F_diss * 2, dCdt_Hplus, dzdt_zsnow)

    return rv


def init_carbonate_system_2(
    export_flux: Flux,
    r_sb: ReservoirGroup,  # Surface box
    r_db: ReservoirGroup,  # deep box
    kwargs: dict,
):
    """Initialize a carbonate system 2 instance.
    Note that the current implmentation assumes that the export flux is
    the total export flux over surface area of the mixed layer, i.e.,
    the sediment area between z0 and zmax

    Parameters
    ----------
    export_flux : Flux
        CaCO3 export flux from the surface box
    r_sb : ReservoirGroup
        ReservoirGroup instance of the surface box
    box r_db : ReservoirGroup
        ReservoirGroup instance of the deep box
    kwargs : dict
        dictionary of keyword value pairs


    """

    AD = r_sb.mo.hyp.area_dz(-abs(kwargs["z0"]), -abs(kwargs["zmax"]))
    s = r_db.swc
    sp = (s.K1, s.K2, s.K1K2, s.KW, s.KB, s.ca2, s.boron, r_sb.DIC.isotopes)
    cp = (
        kwargs["Ksp0"],  # 7
        float(kwargs["kc"]),  # 8
        AD,  # 9
        int(abs(kwargs["zsat0"])),  # 10
        kwargs["I_caco3"],  # 11
        kwargs["alpha"],  # 12
        int(abs(kwargs["zsat_min"])),  # 13
        int(abs(kwargs["zmax"])),  # 14
        int(abs(kwargs["z0"])),  # 15
    )

    ec = ExternalCode(
        name="cs",
        species=r_sb.mo.Carbon.CO2,
        function=carbonate_system_2,
        fname="carbonate_system_2",
        ftype="needs_flux",
        r_s=r_sb,  # source (RG) of CaCO3 flux,
        r_d=r_db,  # sink (RG) of CaCO3 flux,
        function_input_data=[
            export_flux,  # 1
            r_db.DIC,  # 2
            r_db.TA,  # 3
            r_sb.DIC,  # 4
            "Hplus",  # 5
            "zsnow",  # 6
        ],
        function_params=(
            sp,
            cp,
            r_db.mo.area_table,
            r_db.mo.area_dz_table,
            r_db.mo.Csat_table,
        ),
        return_values=[
            {f"F_{r_db.full_name}.DIC": "db_cs2"},
            {f"F_{r_db.full_name}.TA": "db_cs2"},
            {f"R_{r_db.full_name}.Hplus": r_db.swc.hplus},
            {f"R_{r_db.full_name}.zsnow": float(abs(kwargs["zsnow"]))},
        ],
        register=r_db,
    )
    r_db.mo.lpc_f.append(ec.fname)  # list of function to be imported in ode backend

    return ec


def add_carbonate_system_2(**kwargs) -> None:
    """Creates a new carbonate system virtual reservoir
    which will compute carbon species, saturation, compensation,
    and snowline depth, and compute the associated carbonate burial fluxes

    Required keywords:
        r_sb: list of ReservoirGroup objects in the surface layer
        r_db: list of ReservoirGroup objects in the deep layer
        carbonate_export_fluxes: list of flux objects which must match the
        list of ReservoirGroup objects.
        zsat_min = depth of the upper boundary of the deep box
        z0 = upper depth limit for carbonate burial calculations
        typically zsat_min

    Optional Parameters:

        zsat = initial saturation depth (m)
        zcc = initial carbon compensation depth (m)
        zsnow = initial snowline depth (m)
        zsat0 = characteristic depth (m)
        Ksp0 = solubility product of calcite at air-water interface (mol^2/kg^2)
        kc = heterogeneous rate constant/mass transfer coefficient for calcite dissolution (kg m^-2 yr^-1)
        Ca2 = calcium ion concentration (mol/kg)
        pc = characteristic pressure (atm)
        pg = seawater density multiplied by gravity due to acceleration (atm/m)
        I = dissolvable CaCO3 inventory
        co3 = CO3 concentration (mol/kg)
        Ksp = solubility product of calcite at in situ sea water conditions (mol^2/kg^2)

    """

    # list of known keywords
    lkk: dict = {
        "r_db": list,  # list of deep reservoirs
        "r_sb": list,  # list of corresponding surface reservoirs
        "carbonate_export_fluxes": list,
        "zsat": int,
        "zsat_min": int,
        "zcc": int,
        "zsnow": int,
        "zsat0": int,
        "Ksp0": float,
        "kc": float,
        "Ca2": float,
        "pc": (float, int),
        "pg": (float, int),
        "I_caco3": (float, int),
        "alpha": float,
        "zmax": (float, int),
        "z0": (float, int),
        "Ksp": (float, int),
        # "BM": (float, int),
    }
    # provide a list of absolutely required keywords
    lrk: list[str] = [
        "r_db",
        "r_sb",
        "carbonate_export_fluxes",
        "z0",
    ]

    # we need the reference to the Model in order to set some
    # default values.
    reservoir = kwargs["r_db"][0]
    model = reservoir.mo

    # list of default values if none provided
    lod: dict = {
        "r_sb": [],  # empty list
        "zsat": -3715,  # m
        "zcc": -4750,  # m
        "zsnow": -4750,  # m
        "zsat0": -5078,  # m
        "Ksp0": reservoir.swc.Ksp0,  # mol^2/kg^2
        "kc": 8.84 * 1000,  # m/yr converted to kg/(m^2 yr)
        "alpha": 0.6,  # 0.928771302395292, #0.75,
        "pg": 0.103,  # pressure in atm/km
        "pc": 511,  # characteristic pressure after Boudreau 2010
        "I_caco3": 529,  # dissolveable CaCO3 in mol/m^2
        "zmax": -6000,  # max model depth
        "Ksp": reservoir.swc.Ksp_ca,  # mol^2/kg^2
    }
    # make sure all mandatory keywords are present
    __checkkeys__(lrk, lkk, kwargs)
    # add default values for keys which were not specified
    kwargs = __addmissingdefaults__(lod, kwargs)
    # test that all keyword values are of the correct type
    __checktypes__(lkk, kwargs)

    if "zsat_min" not in kwargs:
        kwargs["zsat_min"] = kwargs["z0"]

    r_db = kwargs["r_db"]
    r_sb = kwargs["r_sb"]
    pg = kwargs["pg"]
    pc = kwargs["pc"]
    # test if corresponding surface reservoirs have been defined
    if len(r_sb) == 0:
        raise ValueError(
            "Please update your call to add_carbonate_system_2 and add\
            the list of corresponding surface reservoirs"
        )

    # check if we already have the hypsometry and saturation tables
    if not hasattr(model, "area_table"):
        depth_range = np.arange(0, 6002, 1, dtype=float)  # mbsl
        model.area_table = model.hyp.get_lookup_table(0, -6002)  # area in m^2(z)
        model.area_dz_table = model.hyp.get_lookup_table_area_dz(0, -6002) * -1  # area
        model.Csat_table = (reservoir.swc.Ksp0 / reservoir.swc.ca2) * np.exp(
            (depth_range * pg) / pc
        )

    for i, rg in enumerate(r_db):  # Setup the virtual reservoirs
        if hasattr(rg, "DIC") and hasattr(rg, "TA"):
            rg.swc.update_parameters()
        else:
            raise AttributeError(f"{rg.full_name} must have a TA and DIC reservoir")

        ec = init_carbonate_system_2(
            kwargs["carbonate_export_fluxes"][i],
            r_sb[i],
            r_db[i],
            kwargs,
        )

        register_return_values(ec, rg)
        rg.has_cs2 = True


def get_pco2(SW) -> float:
    """Calculate the concentration of pCO2"""

    dic_c: float = SW.dic
    hplus_c: float = SW.hplus
    k1: float = SW.K1
    k2: float = SW.K2
    co2: NDArrayFloat = dic_c / (1 + (k1 / hplus_c) + (k1 * k2 / (hplus_c**2)))
    pco2: NDArrayFloat = co2 / SW.K0 * 1e6
    return pco2


def calc_pCO2(
    dic,  # see above why no type hints
    hplus,
    SW,
) -> NDArrayFloat:
    """
    Calculate the concentration of pCO2 as a function of DIC,
    H+, K1 and k2 and returns a numpy array containing
    the pCO2 in uatm at each timestep. Calculations are based off
    equations from Follows, 2006. doi:10.1016/j.ocemod.2005.05.004
    dic: Reservoir  = DIC concentrations in mol/kg
    hplus: Reservoir = H+ concentrations in mol/kg
    SW: Seawater = Seawater object for the model
    it is typically used with a DataField object, e.g.
    pco2 = calc_pCO2(dic,h,SW)

     DataField(name = "SurfaceWaterpCO2",
                       associated_with = reservoir_handle,
                       y1_data = pco2,
                       y1_label = r"pCO_{2}",
                       y1_legend = r"pCO_{2}",
                       )

    Author: T. Tsan

    """

    dic_c: NDArrayFloat = dic.c
    hplus_c: NDArrayFloat = hplus.c

    k1: float = SW.K1
    k2: float = SW.K2

    co2: NDArrayFloat = dic_c / (1 + (k1 / hplus_c) + (k1 * k2 / (hplus_c**2)))

    pco2: NDArrayFloat = co2 / SW.K0 * 1e6

    return pco2


def calc_pCO2b(
    dic: NDArrayFloat,
    hplus: NDArrayFloat,
    SW: SeawaterConstants,
) -> NDArrayFloat:
    """
    Same as calc_pCO2, but accepts values/arrays rather than Reservoirs.
    Calculate the concentration of pCO2 as a function of DIC,
    H+, K1 and k2 and returns a numpy array containing
    the pCO2 in uatm at each timestep. Calculations are based off
    equations from Follows, 2006. doi:10.1016/j.ocemod.2005.05.004
    dic:  = DIC concentrations in mol/kg
    hplus: = H+ concentrations in mol/kg
    SW: Seawater = Seawater object for the model
    it is typically used with a DataField object, e.g.
    pco2 = calc_pCO2b(dic,h,SW)

    DataField(name = "SurfaceWaterpCO2",
                      associated_with = reservoir_handle,
                      y1_data = pco2b,
                      y1_label = r"pCO_{2}",
                      y1_legend = r"pCO_{2}",
                      )

    """

    dic_c: NDArrayFloat = dic

    hplus_c: NDArrayFloat = hplus

    k1: float = SW.K1
    k2: float = SW.K2

    co2: NDArrayFloat = dic_c / (1 + (k1 / hplus_c) + (k1 * k2 / (hplus_c**2)))

    pco2: NDArrayFloat = co2 / SW.K0 * 1e6

    return pco2


# define a transform function to display the Hplus concentration as pH
def phc(m: float) -> float:
    """the reservoir class accepts a plot transform. here we use this to
    display the H+ concentrations as pH. After import, you can use it
    with like this in the reservoir definition

     plot_transform_c=phc,

    """
    import numpy as np

    pH = -np.log10(m)
    return pH
